/*********************************************************************/
/*                  pC++/Sage++  Copyright (C) 1993                  */
/*  Indiana University  University of Oregon  University of Rennes   */
/*********************************************************************/

#ifndef LIBSAGEXX_H
#define LIBSAGEXX_H 1

#ifndef __GNUC__
# include <string.h>
#else
# pragma interface
#endif

/* includes the attributes data structure */

////////////////////////////////////////////////////////////////////////////////////////////////////////
//
// Defines the data structure for attributes in sage
// attributes can be used to store any information for any statement, expression, symbol or types nodes
// F. Bodin Indiana July 94.
// 
//
////////////////////////////////////////////////////////////////////////////////////////////////////////

class SgAttribute{
 private:
  // the attribute data;
  int type; // a label;
  void *data;  // the data;
  int dataSize; // the size of the data in bytes to allow data to be copied;
  SgAttribute *next; // to the next attribute of a statements (do that way or not??);
  // link to sage node, allow to go from an attribute to sage stuffs;
  typenode typeNode;   // indicates if SgStatement, SgExpression, ... ptToSage is pointed to;
  void *ptToSage; // pointer to SgStatement, SgExpression, ... ;
  int fileNumber; // the file methods;
// the methods to access the structure of an attributes;  
 public:
  SgAttribute(int t, void *pt, int size, SgStatement &st, int filenum);
  SgAttribute(int t, void *pt, int size, SgSymbol &st, int filenum);
  SgAttribute(int t, void *pt, int size, SgExpression &st, int filenum);
  SgAttribute(int t, void *pt, int size, SgType &st, int filenum);
  ~SgAttribute();
  int getAttributeType();
  void setAttributeType(int t);
  void *getAttributeData();
  void *setAttributeData(void *d);
  int getAttributeSize();
  void setAttributeSize(int s);
  typenode getTypeNode();
  void *getPtToSage(); 
  void  setPtToSage(void *sa);
  void  resetPtToSage();
  void  setPtToSage(SgStatement &st);
  void  setPtToSage(SgSymbol &st);
  void  setPtToSage(SgExpression &st);
  void  setPtToSage(SgType &st);
  SgStatement *getStatement();
  SgExpression *getExpression();
  SgSymbol  *getSgSymbol();
  SgType  *getType();
  int getfileNumber();
  SgAttribute *copy(); 
  SgAttribute *getNext();
  void setNext(SgAttribute *s);
  int listLenght();
  SgAttribute *getInlist(int num);
  void save(FILE *file);
  void save(FILE *file, void (*savefunction)(void *dat,FILE *f));
  
};

///////////////////////////////////////////////////////////////////////////////////////
// The ATTRIBUTE TYPE ALREADY USED
///////////////////////////////////////////////////////////////////////////////////////

#define DEPENDENCE_ATTRIBUTE   -1001
#define INDUCTION_ATTRIBUTE    -1002
#define ACCESS_ATTRIBUTE       -1003
#define DEPGRAPH_ATTRIBUTE     -1004
#define USEDLIST_ATTRIBUTE     -1005
#define DEFINEDLIST_ATTRIBUTE  -1006

#define NOGARBAGE_ATTRIBUTE    -1007
#define GARBAGE_ATTRIBUTE      -1008

// store the annotation expression; it is then visible from the
// garbage collection
#define ANNOTATION_EXPR_ATTRIBUTE   -1009

/**************************************************************
File: libSage++.h
Included in: sage++user.h and libSage++.C

Purpose:It contains all the class definitions and the inline
definitions in Sage++. The start of each class and the start of inlines
in each class are easily identifiable. For example the SgProject class
definition starts with class  SgProject (note the 2 spaces between
class and SgProject) and the comment line preceding the inline
declarations of SgProject is something like // SgProject--inlines.
Sections of the include file are within a #ifdef USER  #endif. Those sections
are included only in sage++user.h and not in libSage++.C. Sections of
the include file are within a #if 0  #endif.  These refer to the unimplemented
portions of Sage++ library.
***************************************************************/

class  SgProject {
  public:
   SgProject(SgProject &);
  SgProject(char * proj_name, int argc, char * argv[]);
   ~SgProject();
   int numberOfFiles();
  SgFile &file(int i);   
   char *fileName(int i); 
   int Fortranlanguage(); 
   int Clanguage();                         
  void addFile(char * dep_file_name);
  void deleteFile(SgFile * file);
};

class  SgFile{
  public:
  PTR_FILE filept;
  SgFile(char* file_name); // the file must exist.
  SgFile(int Language, char* file_name); // for new empty file objects.
   ~SgFile();
   SgFile(SgFile &);
   int languageType();
   void saveDepFile(char *dep_file);
   void unparse(FILE *filedisc);
   void unparsestdout();   

   SgStatement *mainProgram();
  SgStatement *functions(int i);
   int numberOfFunctions();
  SgStatement *getStruct(int i);
   int numberOfStructs();

   SgStatement *firstStatement();
   SgSymbol *firstSymbol();
   SgType *firstType();
   SgExpression *firstExpression();

   SgExpression *SgExpressionWithId(int i);
   SgStatement *SgStatementWithId( int id);
   SgStatement *SgStatementAtLine(int lineno);
   SgSymbol *SgSymbolWithId( int id);
   SgType *SgTypeWithId( int id);
  // for attributes;
  void saveAttributes(char *file);
  void saveAttributes(char *file, void  (*savefunction)(void *dat,FILE *f)); 
  void readAttributes(char *file);
  void readAttributes(char *file, void * (*readfunction)(FILE *f));
  int numberOfAttributes();
  SgAttribute *attribute(int i);
  int SgFile::expressionGarbageCollection(int deleteExpressionNode, int verbose);
};

// Discuss about control parent, BIF structure etc
class  SgStatement{
  public:
  PTR_BFND thebif;
  SgStatement(int variant); 
  SgStatement(PTR_BFND bif);
  SgStatement(int code, SgLabel *lab, SgSymbol *symb, SgExpression *e1, SgExpression *e2, SgExpression *e3);
   SgStatement(SgStatement &);
  // info about statement
   int lineNumber();          // source text line number
   int id();                  // unique id;
   int variant();             // the type of the statement
  SgExpression *expr(int i); // i = 0,1,2 returns the i-th expression.

   int hasSymbol();  // returns TRUE if tmt has symbol, FALSE otherwise
  // returns the symbol field. Used by loop headers to point to the
  // loop variable symbol; Used by function and subroutine headers to
  // point to the function or subroutine name.
  SgSymbol *symbol();        // returns the symbol field.
   char *fileName();
            
   int hasLabel();     // returns 1 if there is a label on the stmt.
  SgLabel *label();          // the label

  // modifying the info.
   void setlineNumber(int n); // change the line number info
   void setId(int n);         // cannot change the id info
   void setVariant(int n);    // change the type of the statement
  void setExpression (int i, SgExpression &e); // change the i-th expression
   void setLabel(SgLabel &l); // change the label
   void setSymbol(SgSymbol &s); // change the symbol

  // Control structure
   SgStatement *lexNext();   // the next statement in lexical order.
   SgStatement *lexPrev();   // the previous stmt in lexical order.
   SgStatement *controlParent(); // the enclosing control statement

   void setLexNext(SgStatement &s); // change the lexical ordering
  void setControlParent(SgStatement &s); // change the control parent.
  void setControlParent(SgStatement *s); // change the control parent.

// Access statement using the tree structure
// Describe BLOB lists here?

   int numberOfChildrenList1();
   int numberOfChildrenList2();
   SgStatement *childList1(int i);
   SgStatement *childList2(int i);
  SgStatement *nextInChildList();

   SgStatement *lastDeclaration(); 
   SgStatement *lastExecutable();  
   SgStatement *lastNodeOfStmt();
   SgStatement *nodeBefore();
   void insertStmtBefore(SgStatement &s);
   void insertStmtBefore(SgStatement &s, SgStatement &cp); 
  void insertStmtAfter(SgStatement &s);
  void insertStmtAfter(SgStatement &s, SgStatement &cp);
   SgStatement *extractStmt();
   SgStatement *extractStmtBody();
   void replaceWithStmt(SgStatement &s);
   void simpledeleteStmt();
  void deleteStmt();
   SgStatement  &copy (void);
   SgStatement  *copyPtr (void);
   SgStatement  &copyOne (void);
   SgStatement  *copyOnePtr (void);
   SgStatement  &copyBlock (void);
   SgStatement  *copyBlockPtr (void);
   int isIncludedInStmt(SgStatement &s);
   void replaceSymbByExp(SgSymbol &symb, SgExpression &exp);
   void replaceSymbBySymb(SgSymbol &symb, SgSymbol &newsymb);
   void replaceSymbBySymbSameName(SgSymbol &symb, SgSymbol &newsymb);
   void replaceTypeInStmt(SgType &old, SgType &newtype);
  char* unparse();
   void unparsestdout();
  void sunparse(char *buffer); //unparsing functions.
   char *comments();      //preceding comment lines.
  void addComment(char *com);
  /* ajm: setComments: set ALL of the node's comments */
   void setComments (char *comments);

  int numberOfComments(); //number of preceeding comments. CAREFUL! 

  int hasAnnotations();   //1 if there are annotations; 0 otherwise

  ~SgStatement();

  // the following functions are added for compiler to look up and build symbol;

  SgSymbol * SgStatement::GetSymbolInScope(SgStatement * scope, char * name);
  SgSymbol * SgStatement::GetSymbolInScope(SgStatement * scope, char * mixstring, char * genname, char * basename); 
  SgSymbol * SgStatement::GetSymbolInScope(SgStatement * scope, char * mixstring, char * genname, char * basename, int counter);
  SgSymbol * SgStatement::GetSymbolInScope(SgStatement * scope, char * mixstring, char * genname, char * basename, int c1, int c2);

  SgSymbol * SgStatement::getSymbolInCurrentScope(char * name);
  SgSymbol * SgStatement::getSymbolInCurrentScope(char * mixstring, char * genname, char * basename);
  SgSymbol * SgStatement::getSymbolInCurrentScope(char * mixstring, char * genname, char * basename, int counter);
  SgSymbol * SgStatement::getSymbolInCurrentScope(char * mixstring, char * genname, char * basename, int c1, int c2);

  // These function must be removed. Doesn't make sense here.
  int IsSymbolInScope(SgSymbol &symb); // TRUE if symbol is in scope
  int IsSymbolReferenced(SgSymbol &symb);
   SgStatement *getScopeForDeclare(); // return where a variable can be declared;

  /////////////// FOR ATTRIBUTES //////////////////////////

  int numberOfAttributes();
  int numberOfAttributes(int type); // of a specified type;
  void *attributeValue(int i); 
  int  attributeType(int i); 
  void *attributeValue(int i,int type); // only considering one type attribute
  void *deleteAttribute(int i); 
  void addAttribute(int type, void *a, int size); // void * can be NULL;
  void addAttribute(int type); //void * is NULL;
  void addAttribute(void *a, int size); //no type specifed;
  void addAttribute(SgAttribute *att);
  SgAttribute *getAttribute(int i);
  SgAttribute *getAttribute(int i,int type);

  //////////// FOR DECL_SPECS (friend, inline, extern, static) ////////////

   void addDeclSpec(int type);   //type should be one of BIT_EXTERN,
                                //BIT_INLINE, _FRIEND, BIT_STATIC
  void clearDeclSpec();        //resets the decl_specs field to zero
   int isFriend();               //returns non-zero if friend modifier set
                                //returns zero otherwise
   int isInline();
   int isExtern();
   int isStatic();
};

class  SgExpression{    
public:
  PTR_LLND thellnd;
  // generic expression class.
  SgExpression(int variant, SgExpression &lhs, SgExpression &rhs,
               SgSymbol &s, SgType &type);

  SgExpression(int variant, SgExpression *lhs, SgExpression *rhs,
 	       SgSymbol *s, SgType *type);
  SgExpression(int variant, SgExpression *lhs, SgExpression *rhs,
 	       SgSymbol *s);
  // for some node in fortran
  SgExpression(int variant,char *str);

  SgExpression(int variant);
  SgExpression(PTR_LLND ll);
   SgExpression(SgExpression &);
   ~SgExpression();
  
   SgExpression *lhs();
   SgExpression *rhs();
  SgExpression *operand(int i);
   int variant();      
   SgType *type(); 
  SgSymbol *symbol();
   int id();
   SgExpression *nextInExprTable();
  
   void setLhs(SgExpression &e);
   void setLhs(SgExpression *e);
   void setRhs(SgExpression &e);
   void setRhs(SgExpression *e);
   void setSymbol(SgSymbol &s);
   void setSymbol(SgSymbol *s);
   void setType(SgType &t);
   void setType(SgType *t);
   void setVariant(int v);
  
   SgExpression &copy();
   SgExpression *copyPtr();
  char *unparse(); 
  void sunparse(char *buffer);
   void unparsestdout();
   SgExpression *IsSymbolInExpression(SgSymbol &symbol);
   void replaceSymbolByExpression(SgSymbol &symbol, SgExpression &expr);
   SgExpression *symbRefs();
   SgExpression *arrayRefs(); 
  int  linearRepresentation(int *coeff, SgSymbol **symb,int *cst, int size);
  SgExpression *normalForm(int n, SgSymbol *s);        
  SgExpression *coefficient(SgSymbol &s); 
  int isInteger();
  int valueInteger();

  void replace(SgExpression & e);

friend SgExpression * operator + ( SgExpression *lhs, SgExpression &rhs); 
friend SgExpression * operator * ( SgExpression *lhs, SgExpression &rhs); 


friend SgExpression &operator + ( SgExpression &lhs, SgExpression &rhs); 
friend SgExpression &operator - ( SgExpression &lhs, SgExpression &rhs); 
friend SgExpression &operator * ( SgExpression &lhs, SgExpression &rhs); 
friend SgExpression &operator / ( SgExpression &lhs, SgExpression &rhs); 
friend SgExpression &operator % ( SgExpression &lhs, SgExpression &rhs); 
friend SgExpression &operator <<( SgExpression &lhs, SgExpression &rhs); 
friend SgExpression &operator >>( SgExpression &lhs, SgExpression &rhs); 
friend SgExpression &operator < ( SgExpression &lhs, SgExpression &rhs); 
friend SgExpression &operator > ( SgExpression &lhs, SgExpression &rhs);
friend SgExpression &operator <= ( SgExpression &lhs, SgExpression &rhs); 
friend SgExpression &operator >= ( SgExpression &lhs, SgExpression &rhs);
friend SgExpression &operator & ( SgExpression &lhs, SgExpression &rhs); 
friend SgExpression &operator | ( SgExpression &lhs, SgExpression &rhs); 
friend SgExpression &operator &&( SgExpression &lhs, SgExpression &rhs); 
friend SgExpression &operator ||( SgExpression &lhs, SgExpression &rhs); 
friend SgExpression &operator +=( SgExpression &lhs, SgExpression &rhs); 
friend SgExpression &operator &=( SgExpression &lhs, SgExpression &rhs); 
friend SgExpression &operator *=( SgExpression &lhs, SgExpression &rhs); 
friend SgExpression &operator /=( SgExpression &lhs, SgExpression &rhs); 
friend SgExpression &operator %=( SgExpression &lhs, SgExpression &rhs); 
friend SgExpression &operator ^=( SgExpression &lhs, SgExpression &rhs); 
friend SgExpression &operator <<=( SgExpression &lhs, SgExpression &rhs); 
friend SgExpression &operator >>=( SgExpression &lhs, SgExpression &rhs);
friend SgExpression &SgAssignOp( SgExpression &lhs, SgExpression &rhs); 
friend SgExpression &SgEqOp( SgExpression &lhs, SgExpression &rhs); 
friend SgExpression &SgNeqOp( SgExpression &lhs, SgExpression &rhs); 
friend SgExpression &SgExprListOp( SgExpression &lhs, SgExpression &rhs); 
friend SgExpression &SgRecRefOp( SgExpression &lhs, SgExpression &rhs); 
friend SgExpression &SgPointStOp( SgExpression &lhs, SgExpression &rhs); 
friend SgExpression &SgScopeOp( SgExpression &lhs, SgExpression &rhs); 
friend SgExpression &SgDDotOp( SgExpression &lhs, SgExpression &rhs); 
friend SgExpression &SgBitNumbOp( SgExpression &lhs, SgExpression &rhs); 

  /////////////// FOR ATTRIBUTES //////////////////////////

  int numberOfAttributes();
  int numberOfAttributes(int type); // of a specified type;
  void *attributeValue(int i); 
  int  attributeType(int i); 
  void *attributeValue(int i,int type); // only considering one type attribute
  void *deleteAttribute(int i); 
  void addAttribute(int type, void *a, int size); // void * can be NULL;
  void addAttribute(int type); //void * is NULL;
  void addAttribute(void *a, int size); //no type specifed;
  void addAttribute(SgAttribute *att);
  SgAttribute *getAttribute(int i);
  SgAttribute *getAttribute(int i,int type);
};

class SgSymbol{
public:
  // basic class contains
  PTR_SYMB thesymb;
  SgSymbol(int variant, char *identifier, SgType &t, SgStatement &scope);
  SgSymbol(int variant, char *identifier, SgType *t, SgStatement *scope);
  SgSymbol(int variant, char *identifier,  SgStatement &scope);
  SgSymbol(int variant, char *identifier,  SgStatement *scope);

  SgSymbol(int variant, char *identifier,   SgType *type, SgStatement *scope, SgSymbol *structsymb, SgSymbol *nextfield );

  SgSymbol(int variant, char *name);
  SgSymbol(int variant);
  SgSymbol(PTR_SYMB symb);
   SgSymbol(SgSymbol &);
  ~SgSymbol();
   int variant();  
   int id();             // unique identifier
   char *identifier();   // the text name for the symbol.
   SgType *type();       // the type of the symbol
   void setType(SgType &t);       // the type of the symbol
   void setType(SgType *t);       // the type of the symbol
   SgStatement *scope(); // the SgControlStatement where defined.
   SgSymbol *next();     // next symbol reference.
   SgSymbol *nextInScope();     // next symbol reference in the same scope.
  SgStatement *declaredInStmt();  // the declaration statement
   SgSymbol &copy();
   SgSymbol &copyLevel1(); // copy also parameters
   SgSymbol &copyLevel2(); // copy parameters, body also
   SgSymbol &copyAcrossFiles(SgStatement &where); // special copy to move things from a file.
  int attributes();    // the Fortran 90 attributes
  void setAttribute(int attribute);
  void removeAttribute(int attribute);
  void declareTheSymbol(SgStatement &st);
   void declareTheSymbolWithParamList
                          (SgStatement &st, SgExpression &parlist);
  SgExpression *makeDeclExpr();
   SgExpression *makeDeclExprWithParamList
                          (SgExpression &parlist);
  SgVarDeclStmt *makeVarDeclStmt();
  SgVarDeclStmt *makeVarDeclStmtWithParamList
                          (SgExpression &parlist);
 
  SgStatement *body(); // the body of the symbol if has one (like, function call, class,...)

  /////////////// FOR ATTRIBUTES //////////////////////////

  int numberOfAttributes();
  int numberOfAttributes(int type); // of a specified type;
  void *attributeValue(int i); 
  int  attributeType(int i); 
  void *attributeValue(int i,int type); // only considering one type attribute
  void *deleteAttribute(int i); 
  void addAttribute(int type, void *a, int size); // void * can be NULL;
  void addAttribute(int type); //void * is NULL;
  void addAttribute(void *a, int size); //no type specifed;
  void addAttribute(SgAttribute *att);
  SgAttribute *getAttribute(int i);
  SgAttribute *getAttribute(int i,int type);
};


/* This code by Andrew Mauer (ajm) */
/* These constants are used by SgType::maskDescriptors() and 
   SgType::getTrueType(). */

const int MASK_NO_DESCRIPTORS = ~0; /* all ones = keep everything */
const int MASK_MOST_DESCRIPTORS =  ( BIT_SIGNED | BIT_UNSIGNED
				     | BIT_LONG | BIT_SHORT
				     | BIT_CONST | BIT_VOLATILE );

const int MASK_ALL_DESCRIPTORS = 0; /* keep nothing */


class SgType{
public:
  PTR_TYPE thetype;
  SgType(int variant);
  SgType(int var, SgExpression *len,SgType *base);
  SgType(int var, SgSymbol *symb);
  SgType(int var, SgSymbol *firstfield, SgStatement *structstmt);
   SgType(PTR_TYPE type);
   SgType(SgType &);
   int variant();
   int id();
   SgSymbol *symbol();
   SgType &copy();
   SgType *copyPtr();
   SgType *next();
   int isTheElementType();
   int equivalentToType(SgType &type);
   int equivalentToType(SgType *type);
   SgType *internalBaseType();
   int hasBaseType();
   SgType *baseType();
   SgExpression *length();

/* This code by Andrew Mauer (ajm) */
/*
  maskDescriptors:

  This routine strips many descriptive type traits which you are probably
  not interested in cloning for variable declarations, etc.

  Returns the getTrueType of the base type being described IF there
  are no descriptors which are not masked out. The following masks
  can be specified as an optional second argument:
        MASK_NO_DESCRIPTORS: Do not mask out anything.
        MASK_MOST_DESCRIPTORS: Only leave in: signed, unsigned, short, long,
	                                      const, volatile.
	MASK_ALL_DESCRIPTORS: Mask out everything. 

  If you build your own mask, you should make sure that the traits
  you want to set out have their bits UN-set, and the rest should have
  their bits set. The complementation (~) operator is a good one to use.

  See libSage++.h, where the MASK_*_DESCRIPTORS variables are defined.
*/

   SgType *maskDescriptors (int mask);


/* This code by Andrew Mauer (ajm) */
/*
  getTrueType:

  Since Sage stores dereferenced pointers as PTR(-1) -> PTR(1) -> BASE_TYPE,
  we may need to follow the chain of dereferencing to find the type
  which we expect.

  This code currently assumes that:
  o If you follow the dereferencing pointer (PTR(-1)), you find another
  pointer type or an array type. 

  We do NOT assume that the following situation cannot occur:
      PTR(-1) -> PTR(-1) -> PTR(1) -> PTR(1) -> PTR(-1) -> PTR(1)

  This means there may be more pointers to follow after we come to
  an initial "equilibrium".

  ALGORITHM:

  T_POINTER:
     [WARNING: No consideration is given to pointers with attributes
     (ls_flags) set. For instance, a const pointer is treated the same
     as any other pointer.]
     
     1. Return the same type we got if it is not a pointer type or
     the pointer is not a dereferencing pointer type.

     2. Repeat { get next pointer , add its indirection to current total }
     until the current total is 0. We have reached an equilibrium, so
     the next type will not necessarily be a pointer type.

     3. Check the next type for further indirection with another call
     to getTrueType.

  T_DESCRIPT:
     Returns the result of maskDescriptors called with the given type and mask.

  T_ARRAY:
     If the array has zero dimensions, we pass over it. This type arose
     for me in the following situation:
          double x[2];
	  x[1] = 0;
     
  T_DERIVED_TYPE:
     If we have been told to follow typedefs, get the type of the
     symbol from which this type is derived from, and continue digging.
     Otherwise return this type.


  HITCHES:
     Some programs may dereference a T_ARRAY as a pointer, so we need
     to be prepared to deal with that.
  */

  SgType *getTrueType (int mask = MASK_MOST_DESCRIPTORS, 
                       int follow_typedefs = 0);

  int numberOfAttributes();
  int numberOfAttributes(int type); // of a specified type;
  void *attributeValue(int i); 
  int  attributeType(int i); 
  void *attributeValue(int i,int type); // only considering one type attribute
  void *deleteAttribute(int i); 
  void addAttribute(int type, void *a, int size); // void * can be NULL;
  void addAttribute(int type); //void * is NULL;
  void addAttribute(void *a, int size); //no type specifed;
  void addAttribute(SgAttribute *att);
  SgAttribute *getAttribute(int i);
  SgAttribute *getAttribute(int i,int type);
};

// SgMakeDeclExp can be called by the user to generate declaration
// expressions from type strings.  it handles all C types.
SgExpression *SgMakeDeclExp(SgSymbol *sym, SgType *t);


class SgLabel{
public:
  PTR_LABEL thelabel;
   SgLabel(PTR_LABEL lab);
   SgLabel(SgLabel &);
   SgLabel(int i);
   int id();
   int getLastLabelVal();
   ~SgLabel();
};

class  SgValueExp: public SgExpression{
  // a value of one of the base types
  // variants: INT_VAL, CHAR_VAL, FLOAT_VAL, 
  //           DOUBLE_VAL, STRING_VAL, COMPLEX_VAL, KEYWORD_VAL
public:
   SgValueExp(int value);
   SgValueExp(char char_val);
   SgValueExp(float float_val);
   SgValueExp(double double_val);
   SgValueExp(char *string_val);
   SgValueExp(double real, double imaginary);
   SgValueExp(SgValueExp &real, SgValueExp &imaginary);
   void setValue(int int_val);
   void setValue(char char_val);
   void setValue(float float_val);
   void setValue(double double_val);
   void setValue(char *string_val);
   void setValue(double real, double im);
   void setValue(SgValueExp &real, SgValueExp & im);
   int intValue();
   char*  floatValue();
   char  charValue();
   char*  doubleValue();
   char * stringValue();
   SgExpression *realValue();
   SgExpression *imaginaryValue();
};

class  SgKeywordValExp: public SgExpression{
public:
   SgKeywordValExp(char *name);
   char *value();
};


class  SgUnaryExp: public  SgExpression{
public:
   SgUnaryExp(PTR_LLND ll);
   SgUnaryExp(int variant, SgExpression & e);
   SgUnaryExp(int variant, int post, SgExpression & e);
   int post();
  SgExpression &operand();
};

class  SgCastExp: public SgExpression{
public:
   SgCastExp(PTR_LLND ll);
   SgCastExp(SgType &t, SgExpression &e);
   SgCastExp(SgType &t);
   ~SgCastExp();
};

// delete [size]  expr
// variant == DELETE_OP
class SgDeleteExp: public SgExpression{
public:
   SgDeleteExp(PTR_LLND ll);
   SgDeleteExp(SgExpression &size, SgExpression &expr);
   SgDeleteExp(SgExpression &expr);
   ~SgDeleteExp();
};

// new typename
// new typename (expr)
// variant == NEW_OP
class SgNewExp: public SgExpression{
public:
   SgNewExp(PTR_LLND ll);
   SgNewExp(SgType &t);
   SgNewExp(SgType &t, SgExpression &e);
#if 0
  SgExpression &numberOfArgs();
  SgExpression &argument(int i);
#endif
  ~SgNewExp();
};

// functions here can use LlndMapping perhaps.
class SgExprIfExp: public SgExpression{
  //  (expr1)? expr2 : expr3
  // variant == EXPR_IF
public:
   SgExprIfExp(PTR_LLND ll);
   SgExprIfExp(SgExpression &exp1,SgExpression &exp2, SgExpression &exp3);
  SgExpression &conditional();
  SgExpression &trueExp();
  SgExpression &falseExp();
   void setConditional(SgExpression &c);
  void setTrueExp(SgExpression &t);
  void setFalseExp(SgExpression &f);
  ~SgExprIfExp();
};

class SgFunctionRefExp: public SgExpression{
   // function_name(formal args)  - for function headers and protytpes.
   // variant = FUNCTION_REF
public:
   SgFunctionRefExp(PTR_LLND ll);
   SgFunctionRefExp(SgSymbol &fun);
   ~SgFunctionRefExp();
   SgSymbol *funName();
   SgExpression *args();
   int numberOfArgs();
   SgExpression *arg(int i);
  SgExpression * AddArg(char *, SgType &);
  void addArg(SgExpression &); // this function is added like the one in SgCallStmt;
};

class SgFunctionCallExp: public SgExpression{
  // function_name(expr1, expr2, ....)
  // variant == FUNC_CALL
public:
   SgFunctionCallExp(PTR_LLND ll);
   SgFunctionCallExp(SgSymbol &fun, SgExpression &paramList);
   SgFunctionCallExp(SgSymbol &fun);
   ~SgFunctionCallExp();
   SgSymbol *funName();
   SgExpression *args();
   int numberOfArgs();
   SgExpression *arg(int i);
   void addArg(SgExpression &arg);
};

class SgFuncPntrExp: public SgExpression{
  // (functionpointer)(expr1,expr2,expr3)
  // variant == FUNCTION_OP
public:
   SgFuncPntrExp(PTR_LLND ll);
   SgFuncPntrExp(SgExpression &ptr);
   ~SgFuncPntrExp();
   SgExpression *funExp();
   void setFunExp(SgExpression &s);
   int numberOfArgs();
   SgExpression *arg(int i);
   void addArg(SgExpression &arg);  // add an argument.
  SgExpression* AddArg(SgSymbol *thefunc, char *name, SgType &); 
              // add a formal parameter
              // to a pointer to a function prototype or parameter.
              // this returns the expression
};

class SgExprListExp: public SgExpression{
        // variant == EXPR_LIST , CONTROL_LIST
public:
   SgExprListExp(PTR_LLND ll);
   SgExprListExp();
   SgExprListExp(SgExpression &ptr);
   ~SgExprListExp();
   int length();
   SgExpression *elem(int i);
   SgExprListExp *next();
   SgExpression *value();
   void setValue(SgExpression &ptr);      
   void append(SgExpression &arg);
   void insertExpAfter(SgExpression & exp);
  void linkToEnd(SgExpression &arg);
};

class  SgRefExp: public SgExpression{
  // Fortran name references
  // variant == CONST_REF, TYPE_REF, INTERFACE_REF
public:
   SgRefExp(PTR_LLND ll);
   SgRefExp(int variant, SgSymbol &s);
   ~SgRefExp();
};

class SgTypeRefExp: public SgExpression{
   // a reference to a type in a c++ template argument 
  public:
     SgTypeRefExp(SgType &t);
     ~SgTypeRefExp();
     SgType *getType();
};

class  SgVarRefExp: public SgExpression{
  // scalar variable reference or non-indexed array reference
  // variant == VAR_REF
public:
   SgVarRefExp (PTR_LLND ll);
   SgVarRefExp(SgSymbol &s);
   SgVarRefExp(SgSymbol *s);
  SgExpression *progatedValue(); // if scalar propogation worked
   ~SgVarRefExp();
};


class  SgThisExp: public SgExpression{
  // variant == THIS_NODE
public:
   SgThisExp (PTR_LLND ll);
   SgThisExp(SgType &t);
   ~SgThisExp();
};

class SgTripletExp : public SgExpression {
  // variant ==DDOT;
public:
   SgTripletExp(SgExpression * item0, SgExpression * item1, SgExpression *item2);
   int numberOfItems();
   SgExpression * item(int i);
};

class  SgArrayRefExp: public SgExpression{
  // an array reference
  // variant == ARRAY_REF
public:
   SgArrayRefExp(PTR_LLND ll);
   SgArrayRefExp(SgSymbol &s);
   SgArrayRefExp(SgSymbol &s, SgExpression &subscripts);
   SgArrayRefExp(SgSymbol &s, SgExpression &sub1,SgExpression &sub2);
  
   SgArrayRefExp(SgSymbol &s, SgExpression &sub1,SgExpression &sub2,SgExpression &sub3);
  
   SgArrayRefExp(SgSymbol &s, SgExpression &sub1,SgExpression &sub2,SgExpression &sub3,SgExpression &sub4);
   ~SgArrayRefExp();
   int numberOfSubscripts();  // the number of subscripts in reference
   SgExpression *subscripts();
   SgExpression *subscript(int i);
   void addSubscript(SgExpression &e);
   void setSymbol(SgSymbol &s);
   int isScalarOnDim(int i);

};

// set NODE _TYPE.
class SgPntrArrRefExp: public SgExpression{
public:
   SgPntrArrRefExp(PTR_LLND ll);
   SgPntrArrRefExp(SgExpression &p);
   SgPntrArrRefExp(SgExpression &p, SgExpression &subscripts);
   SgPntrArrRefExp(SgExpression &p, int n, SgExpression &sub1, SgExpression &sub2);
   SgPntrArrRefExp(SgExpression &p, int n, SgExpression &sub1, SgExpression &sub2, SgExpression &sub3);
   SgPntrArrRefExp(SgExpression &p, int n, SgExpression &sub1, SgExpression &sub2, SgExpression &sub3, SgExpression &sub4);
   ~SgPntrArrRefExp();
   int dimension();  // the number of subscripts in reference
   SgExpression *subscript(int i);
   void addSubscript(SgExpression &e);
   void setPointer(SgExpression &p);
};

class SgPointerDerefExp: public SgExpression{
  // pointer dereferencing
  // variant == DEREF_OP
public:
   SgPointerDerefExp(PTR_LLND ll);
   SgPointerDerefExp(SgExpression &pointerExp);
   ~SgPointerDerefExp();
   SgExpression *pointerExp();
};

class SgRecordRefExp: public SgExpression{
  // a field reference of a structure
  // variant == RECORD_REF
public:
   SgRecordRefExp(PTR_LLND ll);
   SgRecordRefExp(SgSymbol &recordName, char *fieldName);
   SgRecordRefExp(SgExpression &recordExp, char *fieldName);
   ~SgRecordRefExp();
   SgSymbol *fieldName();
   SgSymbol *recordName();
   SgExpression *record();
};


class SgStructConstExp:  public SgExpression{
  // Fortran 90 structure constructor
  // variant == STRUCTURE_CONSTRUCTOR
public:
   SgStructConstExp(PTR_LLND ll);
  // further checks on values need to be done.
   SgStructConstExp(SgSymbol &structName, SgExpression &values);
   SgStructConstExp(SgExpression  &typeRef, SgExpression &values);
   ~SgStructConstExp();
   int numberOfArgs();
   SgExpression *arg(int i);
};

class SgConstExp:  public SgExpression{
public:
   SgConstExp(PTR_LLND ll);
   SgConstExp(SgExpression &values);
   ~SgConstExp();
   int numberOfArgs();
   SgExpression *arg(int i);
};

class SgVecConstExp: public SgExpression{
  // a vector constant of the form: [ expr1, expr2, expr3]
  // variant == VECTOR_CONST
public:
   SgVecConstExp(PTR_LLND ll);
   SgVecConstExp(SgExpression &expr_list);
   SgVecConstExp(int n, SgExpression *components);
   ~SgVecConstExp();
  
   SgExpression *arg(int i); // the i-th term
   int numberOfArgs();
   void setArg(int i, SgExpression &e);
};

class SgInitListExp: public SgExpression{
  // used for initializations.  form: { expr1,expr2,expr3}
  // variant == INIT_LIST
public:
   SgInitListExp(PTR_LLND ll);
   SgInitListExp(SgExpression &expr_list);
   SgInitListExp(int n, SgExpression *components);
   ~SgInitListExp();
  
   SgExpression *arg(int i); // the i-th term
   int numberOfArgs();
   void setArg(int i, SgExpression &e);
};

class SgObjectListExp: public SgExpression{
  // used for EQUIVALENCE, NAMELIST and COMMON statements
  // variant == EQUI_LIST, NAMELIST_LIST, COMM_LIST
public:
   SgObjectListExp(PTR_LLND ll);
   SgObjectListExp(int variant, SgSymbol &object, SgExpression &list);
   SgObjectListExp(int variant,SgExpression &objectRef, SgExpression &list);
   ~SgObjectListExp();
   SgSymbol *object();
   SgExpression *objectRef();
   int listLength();
   SgSymbol *symbol(int i);        //i'th symbol
   SgExpression *objectRef(int i);  //i'th symbol
};


class SgAttributeExp: public SgExpression{
  // Fortran 90 attributes
  // variant ==  PARAMETER_OP, PUBLIC_OP, PRIVATE_OP, ALLOCATABLE_OP,
  // DIMENSION_OP, EXTERNAL_OP, IN_OP, OUT_OP, INOUT_OP, INTRINSIC_OP, 
  // POINTER_OP, OPTIONAL_OP, SAVE_OP, TARGET_OP
public:
   SgAttributeExp(PTR_LLND ll);
   SgAttributeExp(int variant);
   ~SgAttributeExp();
};

class SgKeywordArgExp: public SgExpression{
  // Fortran 90 keyword argument
  // variant == KEYWORD_ARG
public:
   SgKeywordArgExp(PTR_LLND ll);
   SgKeywordArgExp(char *argName, SgExpression &exp);
   ~SgKeywordArgExp();
   SgSymbol *arg();
   SgExpression *value();
};

class SgSubscriptExp: public SgExpression{
  // Fortran 90 vector subscript expression
  // variant == DDOT
public:
   SgSubscriptExp(PTR_LLND ll);
   SgSubscriptExp(SgExpression &exp);
   SgSubscriptExp(SgExpression &lbound, SgExpression &ubound, SgExpression &step);
   SgSubscriptExp(SgExpression &lbound, SgExpression &ubound);
   ~SgSubscriptExp();
  // perhaps this function can use LlndMapping
  SgExpression *lbound();
  SgExpression *ubound();
  SgExpression *step();
};

// add by zheng.gb
class SgIndexOpExp: public SgExpression {
  // variant == INDEX_OP
public:
   SgIndexOpExp(SgVarRefExp &var, SgSubscriptExp &bound);
   ~SgIndexOpExp();
};

class SgUseOnlyExp: public SgExpression{
  // Fortran 90 USE statement ONLY attribute 
  // variant == ONLY_NODE
public:
   SgUseOnlyExp(PTR_LLND ll);
   SgUseOnlyExp(SgExpression &onlyList);
   ~SgUseOnlyExp();
   SgExpression *onlyList();
};

class SgUseRenameExp: public SgExpression{
  // Fortran 90 USE statement renaming
  // variant == RENAME_NODE
public:
   SgUseRenameExp(PTR_LLND ll);
   SgUseRenameExp(SgSymbol &newName, SgSymbol &oldName);
   ~SgUseRenameExp();
   SgSymbol *newName();
   SgSymbol *oldName();
   SgExpression *newNameExp();
   SgExpression *oldNameExp();
};


class SgSpecPairExp: public SgExpression{
  // Fortran default control arguments to Input/Output statements
  // variant == SPEC_PAIR
public:
   SgSpecPairExp(PTR_LLND ll);
   SgSpecPairExp(SgExpression &arg, SgExpression &value);
   SgSpecPairExp(SgExpression &arg);
   SgSpecPairExp(char *arg, char *value);
   ~SgSpecPairExp();
   SgExpression *arg();
   SgExpression *value();
};


//used for do-loop range representation also.
// this form needs to be standardized. 
class SgIOAccessExp: public SgExpression{
  // Fortran index variable bound instantiation
  // variant == IOACCESS
public:
   SgIOAccessExp(PTR_LLND ll);
  // type-checking on bounds needs to be done.
  // Float values are legal in some cases. check manual.
   SgIOAccessExp(SgSymbol &s, SgExpression lbound, SgExpression ubound, SgExpression step);
   SgIOAccessExp(SgSymbol &s, SgExpression lbound, SgExpression ubound);
   ~SgIOAccessExp();
};

class SgImplicitTypeExp: public SgExpression{
  // Fortran index variable bound instantiation
  // variant == IMPL_TYPE
public:
   SgImplicitTypeExp(PTR_LLND ll);
   SgImplicitTypeExp(SgType &type, SgExpression &rangeList);
   ~SgImplicitTypeExp();
   SgType *type();
   SgExpression *rangeList();
   char *alphabeticRange();
};

class SgTypeExp: public SgExpression{
  // Fortran type expression 
  // variant == TYPE_OP
public:
   SgTypeExp(PTR_LLND ll);
   SgTypeExp(SgType &type);
   ~SgTypeExp();
   SgType *type();
};

class SgSeqExp: public SgExpression{
  // Fortran index variable bound instantiation
  // variant == SEQ
public:
   SgSeqExp(PTR_LLND ll);
   SgSeqExp(SgExpression &exp1, SgExpression &exp2);
   ~SgSeqExp();
   SgExpression *front();
   SgExpression *rear();
};

class SgStringLengthExp: public SgExpression{
  // Fortran index variable bound instantiation
  // variant == LEN_OP
public:
   SgStringLengthExp(PTR_LLND ll);
   SgStringLengthExp(SgExpression &length);
   ~SgStringLengthExp();
   SgExpression *length();
};

class SgDefaultExp: public SgExpression {
  // Fortran default node
  // variant == DEFAULT
public:
  SgDefaultExp(PTR_LLND ll);
  SgDefaultExp();
  ~SgDefaultExp();
};

class SgLabelRefExp: public SgExpression{
  // Fortran label reference
  // variant == LABEL_REF
public:
   SgLabelRefExp(PTR_LLND ll);
   SgLabelRefExp(SgLabel &label);
   ~SgLabelRefExp();
   SgLabel *label();
};


class  SgProgHedrStmt: public SgStatement{
  // fortran Program block
  // variant == PROG_HEDR
public:
   SgProgHedrStmt(PTR_BFND bif);
   SgProgHedrStmt(int variant);
   SgProgHedrStmt(SgSymbol &name, SgStatement &Body);
   SgProgHedrStmt(SgSymbol &name);
   SgProgHedrStmt(char *name);
   SgSymbol &name();
   void setName(SgSymbol &symbol); // set program name 

   int numberOfFunctionsCalled();  // the number of functions called
   SgSymbol *calledFunction(int i);// the i-th called function
   int numberOfStmtFunctions();   // the number of statement funcions;
   SgStatement *statementFunc(int i); // the i-th statement function;
   int numberOfEntryPoints();     // the number of entry points;
   SgStatement *entryPoint(int i); // the i-th entry point;
   int numberOfParameters();       // the number of parameters;       
   SgSymbol *parameter(int i);     // the i-th parameter
   void insertParameterAfter(SgSymbol & parameter, SgSymbol & added);
   int numberOfSpecificationStmts();
   SgStatement *specificationStmt(int i);
   int numberOfExecutionStmts();
   SgStatement *executionStmt(int i);
   int numberOfInternalFunctionsDefined();
   SgStatement *internalFunction(int i);
   int numberOfInternalSubroutinesDefined();
   SgStatement *internalSubroutine(int i);
   int numberOfInternalSubProgramsDefined();
   SgStatement *internalSubProgram(int i);
  
#if 0
  SgSymbol &addVariable(SgType &T, char *name); 
                                        //add a declaration for new variable
  
  SgStatement &addCommonBlock(char *blockname, int noOfVars,
                              SgSymbol *Vars); // add a new common block
#endif        
   int isSymbolInScope(SgSymbol &symbol);
   int isSymbolDeclaredHere(SgSymbol &symbol);

   SgSymbol * firstSymbolInScope();
  
  // global analysis data
  
   int numberOfVarsUsed();  // list of used variable access sections
   SgExpression *varsUsed(int i); // i-th var used section descriptor
   int numberofVarsMod();   // list of modifed variable access sections
   SgExpression *varsMod(int i);   // i-th var mod section descriptor
   ~SgProgHedrStmt();
};

class  SgProcHedrStmt: public SgProgHedrStmt{
  // Fortran subroutine
  // variant == PROC_HEDR
public:        
   SgProcHedrStmt(int variant);
   SgProcHedrStmt(SgSymbol &name, SgStatement &Body);
   SgProcHedrStmt(SgSymbol &name);
   SgProcHedrStmt(char *name);
   void AddArg(SgExpression &arg); 
  SgExpression * AddArg(char *name, SgType &t); // returns decl expr created.
  SgExpression * AddArg(char *name, SgType &t, SgExpression &initializer);
   int isRecursive();  // 1 if recursive.;
   int numberOfEntryPoints();      // the number of entry points
                                         // other than the main, 0 for C funcs.
   SgStatement *entryPoint(int i);  // the i-th entry point
  // this is incorrect. Takes only subroutines calls into account.
  // Should be modified to take function calls into account too.
   int numberOfCalls();            // number of calls to this proc.
   SgStatement *call(int i);       // position of the i-th call.  
   ~SgProcHedrStmt();
};      


class  SgProsHedrStmt: public SgProgHedrStmt{
  // Fortran M process
  // variant == PROS_HEDR
public:        
   SgProsHedrStmt();
   SgProsHedrStmt(SgSymbol &name, SgStatement &Body);
   SgProsHedrStmt(SgSymbol &name);
   SgProsHedrStmt(char *name);
   void AddArg(SgExpression &arg);
   int numberOfCalls();            // number of calls to this proc.
   SgStatement *call(int i);       // position of the i-th call.  
   ~SgProsHedrStmt();
};      


class  SgFuncHedrStmt: public SgProcHedrStmt{
  // Fortran and C function.
  // variant == FUNC_HEDR
public:
   SgFuncHedrStmt(SgSymbol &name, SgStatement &Body);
   SgFuncHedrStmt(SgSymbol &name, SgType &type, SgStatement &Body);
   SgFuncHedrStmt(SgSymbol &name, SgSymbol &resultName, SgType &type, SgStatement &Body);
   SgFuncHedrStmt(SgSymbol &name);
   SgFuncHedrStmt(SgSymbol &name, SgExpression *exp);
   SgFuncHedrStmt(char *name);
   ~SgFuncHedrStmt();
  
   SgSymbol *resultName();  // name of result variable.;
  int  setResultName(SgSymbol &symbol); // set name of result variable.;
  
   SgType *returnedType();  // type of returned value
   void setReturnedType(SgType &type);  // set type of returned value
};

class SgClassStmt;

class SgTemplateStmt: public SgStatement{
  // This is a function template or class template
  // in both cases the variant is TEMPLATE_FUNDECL
public:
   SgTemplateStmt(SgExpression *arglist);
  SgExpression * AddArg(char *name, SgType &t); // returns decl expr created.
     // if name == NULL then this is a type reference.
  SgExpression * AddArg(char *name, SgType &t, SgExpression &initializer);
   int numberOfArgs();
   SgExpression *arg(int i);
   SgExpression *argList();
   void addFunction(SgFuncHedrStmt &theTemplateFunc);
   void addClass(SgClassStmt &theTemplateClass);
   SgFuncHedrStmt *isFunction();
   SgClassStmt *isClass();
};

#if 0
class  SgModuleStmt: public SgStatement{
  // Fortran 90 Module statement
  // variant ==  MODULE_STMT
public:
  SgModuleStmt(SgSymbol &moduleName, SgStatement &body);
  SgModuleStmt(SgSymbol &moduleName);
  ~SgModuleStmt();
  
  SgSymbol *moduleName();               // module name 
  void setName(SgSymbol &symbol);        // set module name 
  
  int numberOfSpecificationStmts();
  int numberOfRoutinesDefined();
  int numberOfFunctionsDefined();
  int numberOfSubroutinesDefined();
  
  SgStatement *specificationStmt(int i);
  SgStatement *routine(int i);
  SgStatement *function(int i);
  SgStatement *subroutine(int i);
  
  int isSymbolInScope(SgSymbol &symbol);
  int isSymbolDeclaredHere(SgSymbol &symbol);
  
  SgSymbol &addVariable(SgType &T, char *name); 
                                        //add a declaration for new variable
  
  SgStatement *addCommonBlock(char *blockname, int noOfVars,
                              SgSymbol *Vars); // add a new common block
};

class  SgInterfaceStmt: public SgStatement{
  // Fortran 90 Operator Interface Statement
  // variant == INTERFACE_STMT
public:
  SgInterfaceStmt(SgSymbol &name, SgStatement &body, SgStatement &scope);
  ~SgInterfaceStmt();
  
  SgSymbol *interfaceName();               // interface name if given
  int setName(SgSymbol &symbol);           // set interface name 
  
  int numberOfSpecificationStmts();
  
  SgStatement *specificationStmt(int i);
  
  int isSymbolInScope(SgSymbol &symbol);
  int isSymbolDeclaredHere(SgSymbol &symbol);
};

class  SgBlockDataStmt: public SgStatement{
  // Fortran Block Data statement
  // variant == BLOCK_DATA
public:
  SgBlockDataStmt(SgSymbol &name, SgStatement &body);
  ~SgBlockDataStmt();
  
  SgSymbol *name();  // block data name if given;
  int setName(SgSymbol &symbol);           // set block data name 
  
  int isSymbolInScope(SgSymbol &symbol);
  int isSymbolDeclaredHere(SgSymbol &symbol);
};

#endif


class SgClassStmt: public SgStatement{
  // C++ class statement
  //    class name : superclass_list ElementTypeOf collection_name {
  //            body
  //     } variables_list;
  // variant == CLASS_DECL
public:
   SgClassStmt(int variant);
   SgClassStmt(SgSymbol &name);
   ~SgClassStmt();
   int numberOfSuperClasses();
   SgSymbol *name();
   SgSymbol *superClass(int i);
   void setSuperClass(int i, SgSymbol &symb);
#if 0
  int numberOfVars();            // variables in variables_list
  SgExpression variable(int i);  // i-th variable in variable_list
  SgExpression collectionName(); // if an ElementType class.
  
  // body manipulation functions.
  int numberOfPublicVars();
  int numberOfPrivateVars();
  int numberOfProtectedVars();
  SgSymbol *publicVar(int i);
  SgSymbol *protectedVar(int i);
  SgSymbol *privateVar(int i);
  void addPublicVar(SgSymbol &s);
  void addPrivateVar(SgSymbol &s);
  void addProtectedVar(SgSymbol &s);
  int numberOfPublicFuns();
  int numberOfPrivateFuns();
  int numberOfProtectedFuns();
  SgStatement *publicFun(int i);
  SgStatement *protectedFun(int i);
  SgStatement *privateFun(int i);
  void addPublicFun(SgStatement &s);
  void addPrivateFun(SgStatement &s);
  void addProtectedFun(SgStatement &s);
#endif
};

class SgStructStmt: public SgClassStmt{
  // basic C++ structure
  // struct name ;
  //                   body
  //              } variables_list;
  // variant == STRUCT_DECL
public:
  // consider like a class.
   SgStructStmt();
   SgStructStmt(SgSymbol &name);
   ~SgStructStmt();
  
};


class SgUnionStmt: public SgClassStmt{
  // basic C++ structure
  // union name  {
  //     body
  //    } variables_list;
  // variant == UNION_DECL
public:
  // consider like a class.
   SgUnionStmt();
   SgUnionStmt(SgSymbol &name);
   ~SgUnionStmt();
};

class SgEnumStmt: public SgClassStmt{
  // basic C++ structure
  // enum name  {
  //     body
  //    } variables_list;
  // variant == ENUM_DECL
public:
  // consider like a class.
   SgEnumStmt();
   SgEnumStmt(SgSymbol &name);
   ~SgEnumStmt();
};

class SgCollectionStmt: public SgClassStmt{
  // basic C++ structure
  // collection name ;
  //                    body
  //                } variables_list;
  // variant == COLLECTION_DECL
public:
   SgCollectionStmt();
   SgCollectionStmt(SgSymbol &name);
   ~SgCollectionStmt();
#if 0
  int numberOfElemMethods();
  SgStatement *elementMethod(int i);
  void addElementMethod(SgStatement &s);
#endif
   SgStatement *firstElementMethod();
};

class SgBasicBlockStmt: public SgStatement{
  // in C we have: {  body; }
  // variant == BASIC_BLOCK
public:
   SgBasicBlockStmt();
   ~SgBasicBlockStmt();
};

// ********************* traditional control Structures ************
class  SgForStmt: public SgStatement{
  // for Fortran Do and C for();
  // variant = FOR_NODE
public:
   SgForStmt(SgSymbol &do_var, SgExpression &start, SgExpression &end,
            SgExpression &step, SgStatement &body);
   SgForStmt(SgSymbol *do_var, SgExpression *start, SgExpression *end,
            SgExpression *step, SgStatement *body);
   SgForStmt(SgSymbol &do_var, SgExpression &start, SgExpression &end,
            SgStatement &body);
   SgForStmt(SgExpression &start, SgExpression &end, SgExpression &step,
            SgStatement &body);
  
   SgSymbol doName();                // the name of the loop (for F90.)
   void setDoName(SgSymbol &doName);// sets the name of the loop(for F90)

   SgExpression *start();            
   void setStart(SgExpression &lbound);

   SgExpression *end();
   void setEnd(SgExpression &ubound);

   SgExpression *step();
   void setStep(SgExpression &step);

   SgLabel *endOfLoop();

//SgExpression &bounds();   // bounds are returned as a triplet lb:ub;
//void setBounds(SgTripletOp &bounds); // bounds are passed as a triplet lb:ub;

  // body is returned with control end statement
  //   still attached. 
   SgStatement *body();
  // s is assumed to terminate with a
  //   control end statement.
   void set_body(SgStatement &s);
#if 0
  int replaceBody(SgStatement &s); // new body = s and lex successors.
  
  
  int numberOfInductVars();        // 1 if an induction variable can be found.
  SgSymbol *inductionVar(int i);   // i-th induction variable
  SgExpression *indVarRange(int i); // range of i-th ind. var.
#endif
   int isPerfectLoopNest();
   SgStatement *getNextLoop(); 
   SgStatement *getPreviousLoop();   // returns outer nested loop
   SgStatement *getInnermostLoop();  // returns innermost nested loop
#if 0
  int isLinearLoopNest();          // TRUE if the bound and step of the loops
                                   // in the loop nest are linear expressions
                                 // and use the index variables of the previous
                                // loops of the nest.
#endif
   int isEnddoLoop();            // TRUE if the loop ends with an Enddo
   int convertLoop();            // Convert the loop into a Good loop.
#if 0
  int isAssignLoop();          // TRUE if the body consists only of assignments
  int isAssignIfLoop();        // TRUE if the body consists only of assigments
  //             and conditional statements.
  //high level program transformations. 
  // Most are from SIGMA Toolbox by F.Bodin et al.
  // Semantics can be found in the above reference.
  int tiling_p(int i);     
  int tiling(int i, int tab[]);
  int stripMining(int i);
  SgStatement distributeLoop(int i);
  SgStatement distributeLoopSCC();
  SgStatement loopFusion(SgForStmt &loop);
  SgStatement unrollLoop(int i);
  int interchangeLoops(SgForStmt &loop);
  int interchangeWithLoop(int i);
  int normalized();
  int NormalizeLoop();
  int vectorize();
  int vectorizeNest();
  int ExpandScalar(SgSymbol &symbol, int i);
  int ScalarForwardSubstitution(SgSymbol &symbol);
  int pullStatementToFront(SgStatement &s);
  int pullStatementToEnd(SgStatement &s);
#endif
   ~SgForStmt();
};

// added by zheng.gb
class  SgForallStmt: public SgStatement{
  // for Fortran Do and C for();
  // variant = FORALL_CONSTRUCT_STMT  , FORALL_STMT
public:
   SgForallStmt(SgExpression &forall_header_op , SgStatement &body);
   SgForallStmt(SgExpression &forall_header_op , SgExpression &mask,
            SgStatement &body);

   ~SgForallStmt();
};

class  SgProcessDoStmt: public SgStatement{
  // for Fortran M ProcessDo statement;
  // variant = PROCESS_DO_STAT
public:
   SgProcessDoStmt(SgSymbol &do_var, SgExpression &start,
                         SgExpression &end, SgExpression &step,
                         SgLabel &endofloop, SgStatement &body);
   SgProcessDoStmt(SgSymbol &do_var, SgExpression &start,
                         SgExpression &end, SgLabel &endofloop,
                         SgStatement &body);
   SgProcessDoStmt(SgSymbol &do_var, SgExpression &start,
                         SgExpression &end, SgExpression &step,
                         SgStatement &body);
   SgProcessDoStmt(SgSymbol &do_var, SgExpression &start,
                         SgExpression &end, SgStatement &body);
   SgSymbol doName();
   void setDoName(SgSymbol &doName);
   SgExpression *start();            
   SgExpression *end();
   SgExpression *step();
   SgLabel *endOfLoop();
  // body is returned with control end statement
  //   still attached. 
   SgStatement *body();
  // s is assumed to terminate with a
  //   control end statement.
   void set_body(SgStatement &s);

#if 0
  int replaceBody(SgStatement &s); // new body = s and lex successors.
  
  
  int numberOfInductVars();        // 1 if an induction variable can be found.
  SgSymbol *inductionVar(int i);   // i-th induction variable
  SgExpression *indVarRange(int i); // range of i-th ind. var.
#endif

   int isPerfectLoopNest();
   SgStatement *getNextLoop(); 
   SgStatement *getPreviousLoop();   // returns outer nested loop
   SgStatement *getInnermostLoop();  // returns innermost nested loop
#if 0
  int isLinearLoopNest();          // TRUE if the bound and step of the loops
                                   // in the loop nest are linear expressions
                                 // and use the index variables of the previous
                                // loops of the nest.
#endif
   int isEnddoLoop();            // TRUE if the loop ends with an Enddo
   int convertLoop();            // Convert the loop into a Good loop.
#if 0
  int isAssignLoop();          // TRUE if the body consists only of assignments
  int isAssignIfLoop();        // TRUE if the body consists only of assignments
                               // and conditional statements.
  //high level program transformations. 
  // Most are from SIGMA Toolbox by F.Bodin et al.
  // Semantics can be found in the above reference.
  int tiling_p(int i);     
  int tiling(int i, int tab[]);
  int stripMining(int i);
  SgStatement distributeLoop(int i);
  SgStatement distributeLoopSCC();
  SgStatement loopFusion(SgForStmt &loop);
  SgStatement unrollLoop(int i);
  int interchangeLoops(SgForStmt &loop);
  int interchangeWithLoop(int i);
  int normalized();
  int NormalizeLoop();
  int vectorize();
  int vectorizeNest();
  int ExpandScalar(SgSymbol &symbol, int i);
  int ScalarForwardSubstitution(SgSymbol &symbol);
  int pullStatementToFront(SgStatement &s);
  int pullStatementToEnd(SgStatement &s);
#endif
   ~SgProcessDoStmt();
};


class  SgWhileStmt: public SgStatement{
  // for C while()
  // variant = WHILE_NODE
public:
   SgWhileStmt(int variant);
   SgWhileStmt(SgExpression &cond, SgStatement &body);
   SgExpression *conditional();      // the while test
#if 0
  int numberOfInductVars(); // 1 if an induction variable can be found.
  SgSymbol *inductionVar(int i);    // i-th induction variable
  SgExpression *indVarRange(int i); // range of i-th ind. var.
#endif
   void replaceBody(SgStatement &s); // new body = s and lex successors.
   ~SgWhileStmt();
};

class  SgDoWhileStmt: public SgWhileStmt{
  // For Fortran dowhile().. and C do{....) while();
  // variant = DO_WHILE_NODE
public:
   SgDoWhileStmt(SgExpression &cond, SgStatement &body);
   ~SgDoWhileStmt();
};

// forward reference;
class  SgIfStmt;

class  SgLogIfStmt: public SgStatement{
  // For Fortran logical if  - only one body statement allowed
  // variant == LOGIF_NODE
public:
   SgLogIfStmt(int variant);
   SgLogIfStmt(SgExpression &cond, SgStatement &s);
   SgStatement *body();  // returns reference to first stmt in the body
   SgExpression *conditional();  // the while test
  // check if the statement s is a single statement. 
   void setBody(SgStatement &s); // new body = s 
  // this code won't work, since after the addition false
  //   clause, it should become SgIfThenElse statement.
   void addFalseClause(SgStatement &s); // make it into if-then-else
   SgIfStmt *convertLogicIf();
   ~SgLogIfStmt();
};

class  SgIfStmt: public SgStatement{
  // For Fortran if then else and C if()
  // variant == IF_NODE
public:
   SgIfStmt(int variant);
   SgIfStmt(SgExpression &cond, SgStatement &trueBody, SgStatement &falseBody,
                  SgSymbol &construct_name);
   SgIfStmt(SgExpression &cond, SgStatement &trueBody, SgStatement &falseBody);
   SgIfStmt(SgExpression &cond, SgStatement &trueBody);
   SgStatement *trueBody();      // the first stmt in the True clause
  // SgBlock is needed? 
   SgStatement *trueBody(int i); // i-th stmt in True clause
   SgStatement *falseBody();     // the first stmt in the False
   SgStatement *falseBody(int i);// i-th statement of the body.
   SgExpression *conditional();  // the while test
   SgSymbol *construct_name();
   void replaceTrueBody(SgStatement &s);// new body=s and lex successors.
   void replaceFalseBody(SgStatement &s);//new body=s and lex successors.
   ~SgIfStmt();
};

#if 0
class  SgIfElseIfStmt: public SgIfStmt {
  // For Fortran if then elseif .. elseif ... case
  // variant == ELSEIF_NODE
public:
  SgIfElseIfStmt(SgExpression &condList, SgStatement &blockList,
                 SgSymbol &constructName);
  int numberOfConditionals();       // the number of conditionals
  SgStatement *body(int b);          // block b
  void setBody(int b);              // sets block 
  SgExpression *conditional(int i); // the i-th conditional
  void setConditional(int i);       // sets the i-th conditional
  void addClause(SgExpression &cond, SgStatement &block);
  void removeClause(int b);          // removes block b and it's conditional
  ~SgIfElseIfStmt();
};

 SgIfElseIfStmt::~SgIfElseIfStmt()
{ RemoveFromTableBfnd((void *) this); }

#endif


class  SgArithIfStmt: public SgStatement{
  // For Fortran Arithementic if
  // variant == ARITHIF_NODE
public:
   SgArithIfStmt(int variant);
   SgArithIfStmt(SgExpression &cond, SgLabel &llabel, SgLabel &elabel, SgLabel &glabel);
   SgExpression *conditional();
   void set_conditional(SgExpression &cond);
   SgExpression *label(int i);  // the <, ==, and > goto labels. in order 0->2.
   void setLabel(SgLabel &label);
   ~SgArithIfStmt();
};

class  SgWhereStmt: public SgLogIfStmt{
  // fortran Where stmt
  // variant == WHERE_NODE
public:
   SgWhereStmt(SgExpression &cond, SgStatement &body);
   ~SgWhereStmt();
};

class  SgWhereBlockStmt: public SgIfStmt{
  // fortran Where - Elsewhere stmt
  // variant == WHERE_BLOCK_STMT
public:
  SgWhereBlockStmt(SgExpression &cond, SgStatement &trueBody, SgStatement &falseBody);
  ~SgWhereBlockStmt();
};


class  SgSwitchStmt: public SgStatement{
  // Fortran Case and C switch();
  // variant == SWITCH_NODE
public:
   SgSwitchStmt(SgExpression &selector, SgStatement &caseOptionList,
                      SgSymbol &constructName);
   ~SgSwitchStmt();
   SgExpression *selector();  // the switch selector
   void setSelector(SgExpression &cond);
   int numberOfCaseOptions();       // the number of cases
   SgStatement *caseOption(int i);  // i-th case block
   void addCaseOption(SgStatement &caseOption);
#if 0
  void  deleteCaseOption(int i);
#endif
};

class SgCaseOptionStmt: public SgStatement{
  // Fortran case option statement
  // variant == CASE_NODE
public:
   SgCaseOptionStmt(SgExpression &caseRangeList, SgStatement &body, 
                          SgSymbol &constructName);
   ~SgCaseOptionStmt();
  
   SgExpression *caseRangeList();
   void setCaseRangeList(SgExpression &caseRangeList);
   SgExpression *caseRange(int i);
   void setCaseRange(int i, SgExpression &caseRange);
   SgStatement *body();
   void setBody(SgStatement &body);
};


class  SgExecutableStatement: public  SgStatement{
  // this is really a non-control, non-declaration stmt.
  // no special functions here.
public: 
   SgExecutableStatement(int variant);
};

class SgAssignStmt: public SgExecutableStatement{
  // Fortran assignment Statment
  // variant == ASSIGN_STAT
public:
   SgAssignStmt(int variant);
   SgAssignStmt(SgExpression &lhs, SgExpression &rhs);
   SgExpression *lhs();  // the left hand side
   SgExpression *rhs();  // the right hand side
   void replaceLhs(SgExpression &e); // replace lhs with e
   void replaceRhs(SgExpression &e); // replace rhs with e
#if 0
  SgExpression *varReferenced();
  SgExpression *varUsed();
  SgExpression *varDefined();
#endif
};


class  SgCExpStmt: public SgExecutableStatement{
  // C non-control expression Statment
  // variant == EXPR_STMT_NODE
public:
   SgCExpStmt(SgExpression &exp);
   SgCExpStmt(SgExpression &lhs, SgExpression &rhs);
   SgExpression *expr();  // the expression
   void replaceExpression(SgExpression &e); // replace exp with e
   ~SgCExpStmt();
};

class  SgPointerAssignStmt: public SgAssignStmt{
  // Fortran  pointer assignment statement
  // variant == POINTER_ASSIGN_STAT
public:
   SgPointerAssignStmt(SgExpression lhs, SgExpression rhs);
   ~SgPointerAssignStmt();
};

// heap and nullify statements can be sub-classes 
// of list executable statement class.
class  SgHeapStmt: public SgExecutableStatement{
  // Fortran heap space allocation and deallocation statements
  // variant == ALLOCATE_STMT or DEALLOCATE_STMT 
public:
   SgHeapStmt(int variant, SgExpression &allocationList, SgExpression &statVariable);
   ~SgHeapStmt();
   SgExpression *allocationList();
   void setAllocationList(SgExpression &allocationList);
   SgExpression *statVariable();
   void setStatVariable(SgExpression &statVar);
};

class  SgNullifyStmt: public SgExecutableStatement{
  // Fortran pointer initialization statement
  // variant == NULLIFY_STMT 
public:
   SgNullifyStmt(SgExpression &objectList);
   ~SgNullifyStmt();
   SgExpression *nullifyList();
   void setNullifyList(SgExpression &nullifyList);
};


class  SgContinueStmt: public SgExecutableStatement{
  // variant == CONT_STAT in Fortran and
  // variant == CONTINUE_NODE in C
public:
   SgContinueStmt();
   ~SgContinueStmt();
};

class  SgControlEndStmt: public SgExecutableStatement{
  // the end of a basic block
  // variant == CONTROL_END 
public:
   SgControlEndStmt(int variant);
   SgControlEndStmt();
   ~SgControlEndStmt();
};


class  SgBreakStmt: public SgExecutableStatement{
  // the end of a basic block
  // variant == BREAK_NODE 
public:
   SgBreakStmt();
   ~SgBreakStmt();
};

class  SgCycleStmt: public SgExecutableStatement{
  // the fortran 90 cycle statement
  // variant == CYCLE_STMT
public:
   SgCycleStmt(SgSymbol &symbol);
   SgSymbol *constructName();  // the name of the loop to cycle
   void setConstructName(SgSymbol &constructName);
   ~SgCycleStmt();
};

class  SgReturnStmt: public SgExecutableStatement{
  // the return (expr) node
  // variant == RETURN_NODE//RETURN_STAT
public:
  SgReturnStmt(SgExpression &returnValue);
  SgReturnStmt();
   SgExpression *returnValue();
   void setReturnValue(SgExpression &retVal);
   ~SgReturnStmt();
};


class  SgExitStmt: public SgControlEndStmt{
  // the fortran 90 exit statement
  // variant == EXIT_STMT
public:
   SgExitStmt(SgSymbol &construct_name);
   ~SgExitStmt();
   SgSymbol *constructName();  // the name of the loop to cycle
   void setConstructName(SgSymbol &constructName);
};

class  SgGotoStmt: public SgExecutableStatement{
  // the fortran or C goto
  // variant == GOTO_NODE
public:
   SgGotoStmt(SgLabel &label);
   SgLabel *branchLabel();
#if 0
  SgStatement *target(); //the statement we go to
#endif
   ~SgGotoStmt();
};


class SgLabelListStmt: public SgExecutableStatement{
  // the fortran
  // statements containg a list of labels
public:
  SgLabelListStmt(int variant);
  int numberOfTargets();
  SgExpression *labelList();
  void setLabelList(SgExpression &labelList);
#if 0
  SgStatement *target(int i); //the statement we go to
#endif
};


class  SgAssignedGotoStmt: public SgLabelListStmt{
  // the fortran 
  // variant == ASSGOTO_NODE
public:
  SgAssignedGotoStmt(SgSymbol &symbol, SgExpression &labelList);
  SgSymbol *symbol();
  void setSymbol(SgSymbol &symb);
  ~SgAssignedGotoStmt();
};


class  SgComputedGotoStmt: public SgLabelListStmt{
  // the fortran goto
  // variant == COMGOTO_NODE
public:
   SgComputedGotoStmt(SgExpression &expr, SgLabel &label);
   void addLabel(SgLabel &label);
   SgExpression *exp();
   void setExp(SgExpression &exp);
   ~SgComputedGotoStmt();
};

class  SgStopOrPauseStmt: public SgExecutableStatement{
  // the fortran stop
  // variant == STOP_STAT
public:
  SgStopOrPauseStmt(int variant, SgExpression *expr);
  SgExpression *exp();
  void setExp(SgExpression &exp);
  ~SgStopOrPauseStmt();
};

class  SgCallStmt: public SgExecutableStatement{
  // the fortran call
  // variant == PROC_STAT
public:
  SgCallStmt(SgSymbol &name, SgExpression &args);
  SgCallStmt(SgSymbol &name);
  SgSymbol *name();    // name of subroutine being called
  int numberOfArgs();  // the number of arguement expressions
  void  addArg(SgExpression &arg);
  SgExpression *arg(int i); // the i-th argument expression
  ~SgCallStmt();
  
#if 0
  // global analysis functions
  int numberOfVarsUsed();
  SgExpression *varsUsed(int i);  // i-th region description
  int numberOfVarsMod();
  SgExpression *varsMod(int i);  // i-th region description
#endif
};  


class  SgProsCallStmt: public SgExecutableStatement{
  // the Fortran M process call
  // variant == PROS_STAT
public:
  SgProsCallStmt(SgSymbol &name, SgExprListExp &args);
  SgProsCallStmt(SgSymbol &name);
  SgSymbol *name();    // name of process being called
  int numberOfArgs();  // the number of arguement expressions
  void  addArg(SgExpression &arg);
  SgExprListExp *args();
  SgExpression *arg(int i); // the i-th argument expression
  ~SgProsCallStmt();
};
 
 
class  SgProsCallLctn: public SgExecutableStatement{
  // the Fortran M process call with location
  // variant == PROS_STAT_LCTN
public:
  SgProsCallLctn(SgSymbol &name, SgExprListExp &args, SgExprListExp &lctn);
  SgProsCallLctn(SgSymbol &name, SgExprListExp &lctn);
  SgSymbol *name();    // name of process being called
  int numberOfArgs();  // the number of arguement expressions
  void  addArg(SgExpression &arg);
  SgExprListExp *args();
  SgExpression *arg(int i); // the i-th argument expression
  SgExpression *location();
  ~SgProsCallLctn();
};
 

class  SgProsCallSubm: public SgExecutableStatement{
  // the Fortran M process call with submachine
  // variant == PROS_STAT_SUBM
public:
  SgProsCallSubm(SgSymbol &name, SgExprListExp &args, SgExprListExp &subm);
  SgProsCallSubm(SgSymbol &name, SgExprListExp &subm);
  SgSymbol *name();    // name of process being called
  int numberOfArgs();  // the number of arguement expressions
  void  addArg(SgExpression &arg);
  SgExprListExp *args();
  SgExpression *arg(int i); // the i-th argument expression
  SgExpression *submachine();
  ~SgProsCallSubm();
};
 

class  SgProcessesStmt: public SgStatement{
  // the Fortran M processes statement
  // variant == PROCESSES_STAT
public:
   SgProcessesStmt();
   ~SgProcessesStmt();
};
 

class  SgEndProcessesStmt: public SgStatement{
  // the Fortran M endprocesses statement 
  // variant == PROCESSES_END
public:
   SgEndProcessesStmt();
   ~SgEndProcessesStmt();
};


class SgPortTypeExp: public SgExpression{
  // variant == PORT_TYPE_OP, INPORT_TYPE_OP, or OUTPORT_TYPE_OP
public:
   SgPortTypeExp(SgType &type);
   SgPortTypeExp(SgType &type, SgExpression &ref);
   SgPortTypeExp(int variant, SgExpression &porttype);
   ~SgPortTypeExp();
   SgType *type();
   int numberOfRef();
   SgExpression *ref(); // return a ref or a port type
   SgPortTypeExp *next();
};
 
 
class SgInportStmt: public SgStatement
{
  // the Fortran M inport statement
  // variant == INPORT_DECL
public:
   SgInportStmt(SgExprListExp &name);
   SgInportStmt(SgExprListExp &name, SgPortTypeExp &porttype);
   ~SgInportStmt();
   void addname(SgExpression &name);
   int numberOfNames();
   SgExprListExp *names();
   SgExpression *name(int i);
   void addporttype(SgExpression &porttype);
   int numberOfPortTypes();
   SgPortTypeExp *porttypes();
   SgPortTypeExp *porttype(int i);
};

 
class SgOutportStmt: public SgStatement{
  // the Fortran M outport statement
  // variant == OUTPORT_DECL
public:
   SgOutportStmt(SgExprListExp &name);
   SgOutportStmt(SgExprListExp &name, SgPortTypeExp &porttype);
   ~SgOutportStmt();
   void addname(SgExpression &name);
   int numberOfNames();
   SgExprListExp *names();
   SgExpression *name(int i);
   void addporttype(SgExpression &porttype);
   int numberOfPortTypes();
   SgPortTypeExp *porttypes();
   SgPortTypeExp *porttype(int i);
};

 
class SgChannelStmt: public SgStatement{
  // the Fortran M channel statement
  // variant == CHANNEL_STAT
public:
   SgChannelStmt(SgExpression &outport, SgExpression &inport);
   SgChannelStmt(SgExpression &outport, SgExpression &inport,
                       SgExpression &io_or_err);
   SgChannelStmt(SgExpression &outport, SgExpression &inport,
                       SgExpression &iostore, SgExpression &errlabel);
   ~SgChannelStmt();
   SgExpression *outport();
   SgExpression *inport();
   SgExpression *ioStore();
   SgExpression *errLabel();
};
 

class SgMergerStmt: public SgStatement{
  // the Fortran M merger statement
  // variant == MERGER_STAT
public:
   SgMergerStmt(SgExpression &outport, SgExpression &inport);
   SgMergerStmt(SgExpression &outport, SgExpression &inport,
                     SgExpression &io_or_err);
   SgMergerStmt(SgExpression &outport, SgExpression &inport,
                     SgExpression &iostore, SgExpression &errlabel);
   ~SgMergerStmt();
   void addOutport(SgExpression &outport);
   void addIoStore(SgExpression &iostore); //can't add it before outports
   void addErrLabel(SgExpression &errlabel); //can't add it before iostore
   int numberOfOutports();
   SgExpression *outport(int i);
   SgExpression *inport();
   SgExpression *ioStore();
   SgExpression *errLabel();
};


class SgMoveportStmt: public SgStatement{
  // the Fortran M moveport statement
  // variant == MOVE_PORT
public:
   SgMoveportStmt(SgExpression &fromport, SgExpression &toport);
   SgMoveportStmt(SgExpression &fromport, SgExpression &toport,
                        SgExpression &io_or_err);
   SgMoveportStmt(SgExpression &fromport, SgExpression &toport,
                        SgExpression &iostore, SgExpression &errlabel);
   ~SgMoveportStmt();
   SgExpression *fromport();
   SgExpression *toport();
   SgExpression *ioStore();
   SgExpression *errLabel();
};
 

class SgSendStmt: public SgStatement{
  // the Fortran M send statement
  // variant == SEND_STAT
public:
   SgSendStmt(SgExpression &control, SgExprListExp &argument);
   SgSendStmt(SgExpression &outport, SgExprListExp &argument, SgExpression &io_or_err);
   SgSendStmt(SgExpression &outport, SgExprListExp &argument, SgExpression &iostore, SgExpression &errlabel);
   ~SgSendStmt();
   void addOutport(SgExpression &outport);
   void addIoStore(SgExpression &iostore); //can't add it before outports
   void addErrLabel(SgExpression &errlabel); //can't add it before iostore
   void addArgument(SgExpression &argument);
   int numberOfOutports();
   int numberOfArguments();
   SgExpression *controls();
   SgExpression *outport(int i);
   SgExprListExp *arguments();
   SgExpression *argument(int i);
   SgExpression *ioStore();
   SgExpression *errLabel();
};


class SgReceiveStmt: public SgStatement{
  // the Fortran M receive statement
  // variant == RECEIVE_STAT
public:
   SgReceiveStmt(SgExpression &control, SgExprListExp &argument);
   SgReceiveStmt(SgExpression &inport, SgExprListExp &argument,
                       SgExpression &e1);
   SgReceiveStmt(SgExpression &inport, SgExprListExp &argument,
                       SgExpression &e1, SgExpression &e2);
   SgReceiveStmt(SgExpression &inport, SgExprListExp &argument,
                       SgExpression &e1, SgExpression &e2, SgExpression &e3);
   ~SgReceiveStmt();
   void addInport(SgExpression &inport);
   void addIoStore(SgExpression &iostore);//can't add it before inports
   void addErrLabel(SgExpression &errlabel);//can't add it before iostore
   void addEndLabel(SgExpression &endlabel);//can't add it before errlabel
   void addArgument(SgExpression &argument);
   int numberOfInports();
   int numberOfArguments();
   SgExpression *controls();
   SgExpression *inport(int i);
   SgExprListExp *arguments();
   SgExpression *argument(int i);
   SgExpression *ioStore();
   SgExpression *errLabel();
   SgExpression *endLabel();
};



class SgEndchannelStmt: public SgStatement{
  // the Fortran M endchannel statement
  // variant == ENDCHANNEL_STAT
public:
   SgEndchannelStmt(SgExpression &outport);
   SgEndchannelStmt(SgExpression &outport, SgExpression &io_or_err);
   SgEndchannelStmt(SgExpression &outport, SgExpression &iostore,
                          SgExpression &errlabel);
   ~SgEndchannelStmt();
   void addOutport(SgExpression &outport);
   void addIoStore(SgExpression &iostore);//can't add it before outports
   void addErrLabel(SgExpression &errlabel);//can't add it before iostore
   int numberOfOutports();
   SgExpression *controls();
   SgExpression *outport(int i);
   SgExpression *ioStore();
   SgExpression *errLabel();
};


class SgProbeStmt: public SgStatement{
  // the Fortran M probe statement
  // variant == PROBE_STAT
public:
   SgProbeStmt(SgExpression &inport);
   SgProbeStmt(SgExpression &inport, SgExpression &e1);
   SgProbeStmt(SgExpression &inport, SgExpression &e1,
                     SgExpression &e2);
   SgProbeStmt(SgExpression &inport, SgExpression &e1,
                          SgExpression &e2, SgExpression &e3);
   ~SgProbeStmt();
   void addInport(SgExpression &inport);
   void addIoStore(SgExpression &iostore);//can't add before inports
   void addErrLabel(SgExpression &errlabel);//can't add before iostore
   void addEmptyStore(SgExpression &endlabel);//can't add before errlabel
   int numberOfInports();
   SgExpression *controls();
   SgExpression *inport(int i);
   SgExpression *ioStore();
   SgExpression *errLabel();
   SgExpression *emptyStore();
};


class SgProcessorsRefExp: public SgExpression{
  // variant == PROCESSORS_REF
public:
   SgProcessorsRefExp(PTR_LLND ll);
   SgProcessorsRefExp();
   SgProcessorsRefExp(SgExpression &subscripts);
   SgProcessorsRefExp(SgExpression &sub1,SgExpression &sub2);
 
   SgProcessorsRefExp(SgExpression &sub1,SgExpression &sub2,
                            SgExpression &sub3);
 
   SgProcessorsRefExp(SgExpression &sub1,SgExpression &sub2,
                            SgExpression &sub3,SgExpression &sub4);
   ~SgProcessorsRefExp();
   int numberOfSubscripts();  // the number of subscripts in reference
   SgExpression *subscripts();
   SgExpression *subscript(int i);
   void addSubscript(SgExpression &e);
};


class SgControlExp: public SgExpression{
  //parent of INPORT_NAME, OUTPORT_NAME, FROMPORT_NAME, TOPORT_NAME
  //          IOSTAT_STORE, EMPTY_STORE, ERR_LABEL, END_LABEL
public:
   SgControlExp(int variant);
   ~SgControlExp();
   SgExpression *exp();
};


class SgInportExp: public SgControlExp{
  // variant == INPORT_NAME
public:
   SgInportExp(SgExprListExp &exp);
   ~SgInportExp();
};


class SgOutportExp: public SgControlExp{
  // variant == OUTPORT_NAME
public:
   SgOutportExp(SgExprListExp &exp);
   ~SgOutportExp();
};


class SgFromportExp: public SgControlExp{
  // variant == FROMPORT_NAME
public:
   SgFromportExp(SgExprListExp &exp);
   ~SgFromportExp();
};


class SgToportExp: public SgControlExp{
  // variant == TOPORT_NAME
public:
   SgToportExp(SgExprListExp &exp);
   ~SgToportExp();
};


class SgIO_statStoreExp: public SgControlExp{
  // variant == IOSTAT_STORE
public:
   SgIO_statStoreExp(SgExprListExp &exp);
   ~SgIO_statStoreExp();
};


class SgEmptyStoreExp: public SgControlExp{
  // variant == EMPTY_STORE
public:
   SgEmptyStoreExp(SgExprListExp &exp);
   ~SgEmptyStoreExp();
};


class SgErrLabelExp: public SgControlExp{
  // variant == ERR_LABEL
public:
   SgErrLabelExp(SgExprListExp &exp);
   ~SgErrLabelExp();
};


class SgEndLabelExp: public SgControlExp{
  // variant == END_LABEL
public:
   SgEndLabelExp(SgExprListExp &exp);
   ~SgEndLabelExp();
};


class SgDataImpliedDoExp: public SgExpression{
  // variant == DATA_IMPL_DO
public:
   SgDataImpliedDoExp(SgExprListExp &dlist, SgSymbol &iname,
                            SgExprListExp &ilist);
   ~SgDataImpliedDoExp();
   void addDataelt(SgExpression &data);
   void addIconexpr(SgExpression &icon);
   SgSymbol *iname();
   int numberOfDataelt();
   SgExprListExp *dataelts();
   SgExprListExp *iconexprs(); /* only the first 3 elements in the
                                        iconexpr list are useful. They represent
                                        the initial value, the limit, and the 
                                        increment of the implied do expression
                                        respectively */
   SgExpression *dataelt(int i);
   SgExpression *init();
   SgExpression *limit();
   SgExpression *increment();
};
 

class SgDataEltExp: public SgExpression{
  // variant == DATA_ELT
public:
   SgDataEltExp(SgExpression &dataimplieddo);
   SgDataEltExp(SgSymbol &name, SgExpression &datasubs,
                      SgExpression &datarange);
   ~SgDataEltExp();
   SgExpression *dataimplieddo();
   SgSymbol *name();
   SgExpression *datasubs();
   SgExpression *datarange();
};


class SgDataSubsExp: public SgExpression{
  // variant == DATA_SUBS
public:
   SgDataSubsExp(SgExprListExp &iconexprlist);
   ~SgDataSubsExp();
   SgExprListExp *iconexprlist();
};


class SgDataRangeExp: public SgExpression{
  // variant == DATA_RANGE
public:
   SgDataRangeExp(SgExpression &iconexpr1, SgExpression &iconexpr2);
   ~SgDataRangeExp();
   SgExpression *iconexpr1();
   SgExpression *iconexpr2();
};


class SgIconExprExp: public SgExpression{
  // variant == ICON_EXPR
public:
   SgIconExprExp(SgExpression &expr);
   ~SgIconExprExp();
   SgExpression *expr();
};


class SgImpliedDoExp: public SgExpression {
  // variant == IMPLIED_DO;
public:
  SgImpliedDoExp();
  SgSymbol * index();
  int numberOfTriplet();
  SgExpression * triplet(int);
  SgExpression * list();
  ~SgImpliedDoExp();
};

class SgIOStmt: public SgExecutableStatement{
  // fortran input/output and their control statements
  // abstract class
public:
   SgIOStmt(int variant);
};

class SgInputOutputStmt: public SgIOStmt{
  // fortran input and output statements
  // variant = READ_STAT, WRITE_STATE, PRINT_STAT
public:
   SgInputOutputStmt(int variant, SgExpression &specList, SgExpression &itemList);
   SgExpression *specList();
   void setSpecList(SgExpression &specList);
   SgExpression *itemList();
   void setItemList(SgExpression &itemList);
   ~SgInputOutputStmt();
};

class SgIOControlStmt: public SgExecutableStatement{
  // fortran input/output control and editing statements
  // variant = OPEN_STAT, CLOSE_STAT, INQUIRE_STAT, BACKSPACE_STAT,
  // REWIND_STAT, ENDFILE_STAT, FORMAT_STAT
public:
  SgIOControlStmt(int variant, SgExpression &controlSpecifierList);
   SgExpression *controlSpecList();
   void setControlSpecList(SgExpression &controlSpecList);
   ~SgIOControlStmt();
};

// ******************** Declaration Nodes ***************************

class  SgDeclarationStatement: public SgStatement{
  // Declaration class
  // abstract class
public:
   SgDeclarationStatement(int variant);
   ~SgDeclarationStatement();
  
   SgExpression *varList();
   int numberOfVars();
   SgExpression *var(int i);
   void deleteVar(int i);
   void deleteTheVar(SgExpression &var);
   void addVar(SgExpression &exp);
};

class  SgVarDeclStmt: public SgDeclarationStatement{
  // Declaration Statement
  // variant == VAR_DECL
public:
  // varRefValList is a list of low-level nodes of
  // variants VAR_REFs or ARRAY_REFs or ASSIGN_OPs
   SgVarDeclStmt(SgExpression &varRefValList, SgExpression &attributeList, SgType &type);
   SgVarDeclStmt(SgExpression &varRefValList, SgType &type);
   SgVarDeclStmt(SgExpression &varRefValList);
   ~SgVarDeclStmt();
   SgType *type();  // the type;
   int numberOfAttributes(); // the number of F90 attributes;
  // the attributes are: PARAMETER_OP | PUBLIC_OP |
  //    PRIVATE_OP | ALLOCATABLE_OP | EXTERNAL_OP |
  //    OPTIONAL_OP | POINTER_OP | SAVE_OP TARGET_OP
#if 0
  SgExpression &attribute(int i);
#endif
   int numberOfSymbols();  // the number of variables declared;        
   SgSymbol *symbol(int i);
  
   void deleteSymbol(int i);
   void deleteTheSymbol(SgSymbol &symbol);
   SgExpression *initialValue(int i);  // the initial value ofthe i-th variable
  SgExpression *completeInitialValue(int i); // The complete ASSGN_OP
					// expression of the initial value *BW* from M. Golden
  void setInitialValue(int i, SgExpression &initVal); // sets the initial value ofthe i-th variable
  // an alternative way to initialize variables. The low-level node (VAR_REF or ARRAY_REF) is
  // replaced by a ASSIGN_OP low-level node.
  void clearInitialValue(int i);  // removes initial value of the i-th declaration 
};


class  SgIntentStmt: public SgDeclarationStatement{
  // the Fortran M Intent Statement
  // variant == INTENT_STMT
public:
   SgIntentStmt(SgExpression &varRefValList, SgExpression &attribute);
   ~SgIntentStmt();
   int numberOfArgs();  // the number of arguement expressions
   void  addArg(SgExpression &arg);
   SgExpression *args();
   SgExpression *arg(int i); // the i-th argument expression
   SgExpression *attribute();
};


class  SgVarListDeclStmt: public SgDeclarationStatement{
  // Declaration Statement
  // variant == OPTIONAL_STMT, SAVE_STMT, PUBLIC_STMT,
  // PRIVATE_STMT, EXTERNAL_STAT, INTRINSIC_STAT, DIM_STAT, 
  // ALLOCATABLE_STAT, POINTER_STAT, TARGET_STAT, MODULE_PROC_STMT,
  // PROCESSORS_STAT (for Fortran M processors statement)
public:
  SgVarListDeclStmt(int variant, SgExpression &symbolRefList);
  SgVarListDeclStmt(int variant, SgSymbol &symbolList, SgStatement &scope);

   ~SgVarListDeclStmt();
  
   int numberOfSymbols();
   SgSymbol *symbol(int i);
   void appendSymbol(SgSymbol &symbol);
   void deleteSymbol(int i);
   void deleteTheSymbol(SgSymbol &symbol);
};


class SgStructureDeclStmt: public SgDeclarationStatement{
  // Fortran 90 structure declaration statement
  // variant == STRUCT_DECL
public:
  SgStructureDeclStmt(SgSymbol &name, SgExpression &attributes, SgStatement &body);
  ~SgStructureDeclStmt();
  
#if 0
  int isPrivate();
  int isPublic();
  int isSequence();
#endif
};

class SgNestedVarListDeclStmt: public SgDeclarationStatement{
  // Declaration statement
  // variant == NAMELIST_STAT, EQUI_STAT, COMM_STAT,
  //            and  PROS_COMM for Fortran M
  // These statements have the format of a list of variable lists. For example,
  // EQUIVALENCE (A, C, D), (B, G, F), ....
public:
  SgNestedVarListDeclStmt(int variant, SgExpression &listOfVarList);
  // varList must be of low-level variant appropriate to variant. For example,
  // if the variant is COMM_STAT, listOfVarList must be of variant COMM_LIST.
  ~SgNestedVarListDeclStmt();
  
  SgExpression *lists();
  int numberOfLists();
  SgExpression *list(int i);
#if 0
  SgExpression *leadingVar(int i);
#endif
  void addList(SgExpression &list);
  void  addVarToList(SgExpression &varRef);
  void deleteList(int i);
  void deleteTheList(SgExpression &list);
  void deleteVarInList(int i, SgExpression &varRef);
  void deleteVarInTheList(SgExpression &list, SgExpression &varRef);
};

class SgParameterStmt: public SgDeclarationStatement{      
  // Fortran constants declaration statement
  // variant = PARAM_DECL
public:
  SgParameterStmt(SgExpression &constants, SgExpression &values);
  SgParameterStmt(SgExpression &constantsWithValues);
  ~SgParameterStmt();
  
  int numberOfConstants();    // the number of constants declared
  
  SgSymbol *constant(int i);  // the i-th variable
  SgExpression *value(int i); // the value of i-th variable
  
  void addConstant(SgSymbol &constant, SgExpression &value);
  void deleteConstant(int i);
  void deleteTheConstant(SgSymbol &constant);
};

class SgImplicitStmt: public SgDeclarationStatement{      
  // Fortran implicit type declaration statement
  // variant = IMPL_DECL
public:
  SgImplicitStmt(SgExpression &implicitLists);
  ~SgImplicitStmt();
  
  int numberOfImplicitTypes();  // the number of implicit types declared;
  SgType *implicitType(int i); // the i-th implicit type
  SgExpression *implicitRangeList(int i) ;
  void  appendImplicitNode(SgExpression &impNode);
#if 0
  void  addImplicitType(SgType Type, char alphabet[]);
  int deleteImplicitItem(int i);
  int deleteTheImplicitItem(SgExpression &implicitItem);
#endif
};
#if 0
class SgUseStmt: public SgDeclarationStatement{
  // Fortran 90 module usuage statement
  // variant = USE_STMT
public:
  SgUseStmt(SgSymbol &moduleName, SgExpression &renameList, SgStatement &scope);
  // renameList must be a list of low-level nodes of variant RENAME_NODE
  ~SgUseStmt();
  
  int isOnly();
  SgSymbol *moduleName();
  void setModuleName(SgSymbol &moduleName);
  int numberOfRenames();
  SgExpression *renameNode(int i);
  void  addRename(SgSymbol &localName, SgSymbol &useName);
  void  addRenameNode(SgExpression &renameNode);
  void  deleteRenameNode(int i);
  void deleteTheRenameNode(SgExpression &renameNode);
};




class  SgStmtFunctionStmt: public SgDeclarationStatement{
  // Fortran statement function declaration
  // variant == STMTFN_DECL
public:        
  SgStmtFunctionStmt(SgSymbol &name, SgExpression &args, SgStatement Body);
  ~SgStmtFunctionStmt();
  SgSymbol *name();
  void setName(SgSymbol &name);
  SgType *type();
  int numberOfParameters();       // the number of parameters
  SgSymbol *parameter(int i);     // the i-th parameter
};      

class  SgMiscellStmt: public SgDeclarationStatement{
  // Fortran 90 simple miscellaneous statements
  // variant == CONTAINS_STMT, PRIVATE_STMT, SEQUENCE_STMT
public:        
  SgMiscellStmt(int variant);
  ~SgMiscellStmt();
};      

#endif

class  SgInheritStmt : public SgStatement {
  // for HPF :  INHERIT_DECL ;
public:
  SgInheritStmt();
  int numberOfVars();
  SgExpression * var(int);

};

//
//
// More stuffs for types and symbols
//
//


class SgVariableSymb: public SgSymbol{
  // a variable
  // variant = VARIABLE_NAME
public:
   SgVariableSymb(char *identifier, SgType &t, SgStatement &scope);
   SgVariableSymb(char *identifier, SgType *t, SgStatement *scope);
   SgVariableSymb(char *identifier, SgType &t);
   SgVariableSymb(char *identifier,  SgStatement &scope);
   SgVariableSymb(char *identifier,  SgStatement *scope);
   SgVariableSymb(char *identifier);
   ~SgVariableSymb();

  /* This function allocates and returns a new variable reference
     expression to this symbol. (ajm) */
   SgVarRefExp *varRef (void);

#if 0
  int isAttributeSet(int attribute);
  void setAttribute(int attribute);
  
  int numberOfUses();            // number of uses.
  SgStatement  *useStmt(int i);  // statement where i-th use occurs
  SgExpression *useExpr(int i);  // expression where i-th use occurs
  int numberOfDefs();
#endif
};

class SgConstantSymb: public SgSymbol{
  // a symbol for a constant object
  // variant == CONST_NAME 
public:
   SgConstantSymb(char *identifier, SgStatement &scope, 
                 SgExpression &value);
   ~SgConstantSymb();
   SgExpression *constantValue();
};


class SgFunctionSymb: public SgSymbol{
  // a subroutine, function or main program
  // variant == PROGRAM_NAME, PROCEDURE_NAME, or FUNCTION_NAME
public:
  SgFunctionSymb(int variant);
  SgFunctionSymb(int variant, char *identifier, SgType &t, 
                 SgStatement &scope);
  ~SgFunctionSymb();
  void addParameter(int, SgSymbol &parameters);
  void insertParameter(int position, SgSymbol &symb);
  int numberOfParameters();
  SgSymbol *parameter(int i); 
  SgSymbol *result();
  void  setResult(SgSymbol &symbol);
#if 0
  int isRecursive();
  int setRecursive();
#endif
};


class SgMemberFuncSymb: public SgFunctionSymb{
  // a member function for a class or struct or collection
  // variant = MEMBER_FUNC
  // may be either MEMB_PRIVATE, MEMB_PUBLIC,
  // MEMP_METHOELEM or MEMB_PROTECTED
public:
   SgMemberFuncSymb(char *identifier, SgType &t, SgStatement &cla,
                          int status);
   ~SgMemberFuncSymb();
#if 0
  int status();       
  int isVirtual();       // 1 if virtual.
#endif
   int isMethodOfElement();
   SgSymbol *className();
   void setClassName(SgSymbol &symb);
};

class SgFieldSymb: public SgSymbol{
  // a field in an enum or in a struct.
  // variant == ENUM_NAME or FIELD_NAME
public:
  // no check is made to see if the field "identifier"
  //   already exists in the structure. 
   SgFieldSymb(char *identifier, SgType &t, SgSymbol &structureName);
   ~SgFieldSymb();
   int offset();        // position in the structure
   SgSymbol *structureName();  // parent structure
   SgSymbol *nextField();
   int isMethodOfElement();
#if 0
  int isPrivate();
  int isSequence();
  void  setPrivate();
  void setSequence();
#endif
};

class SgClassSymb: public SgSymbol{
  // the class, union, struct and collection type.
  // variant == CLASS_NAME, UNION_NAME, STRUCT_NAME or COLLECTION_NAME
public:
   SgClassSymb(int variant, char *name, SgStatement &scope);
   ~SgClassSymb();
   int numberOfFields();
   SgSymbol *field(int i);
};

#if 0
class SgTypeSymb: public SgSymbol{
  // a C typedef.  the type() function returns the base type.
  // variant == TYPE_NAME
public:
  SgTypeSymb(char *name, SgType &baseType);
  SgType &baseType();
  ~SgTypeSymb();
};

#endif


class SgLabelSymb: public SgSymbol{
  // a C label name
  // variant == LABEL_NAME
public:
   SgLabelSymb(char *name);
   ~SgLabelSymb();
};


class SgLabelVarSymb: public SgSymbol{
  // a Fortran label variable for an assigned goto stmt
  // variant == LABEL_NAME
public:
   SgLabelVarSymb(char *name, SgStatement &scope);
   ~SgLabelVarSymb();
};

class SgExternalSymb: public SgSymbol{
  // for fortran external statement
  // variant == ROUTINE_NAME
public:
   SgExternalSymb(char *name, SgStatement &scope);
   ~SgExternalSymb();
};

class SgConstructSymb: public SgSymbol{
  // for fortran statement with construct names
  // variant == CONSTRUCT_NAME
public:
   SgConstructSymb(char *name, SgStatement &scope);
   ~SgConstructSymb();
};

// A lot of work needs to be done on this class.
class SgInterfaceSymb: public SgSymbol{
  // for fortran interface statement
  // variant == INTERFACE_NAME
public:
   SgInterfaceSymb(char *name, SgStatement &scope);
   ~SgInterfaceSymb();
};

// A lot of work needs to be done on this class.
class SgModuleSymb: public SgSymbol{
  // for fortran module statement
  // variant == MODULE_NAME
public:
   SgModuleSymb(char *name);
   ~SgModuleSymb();
};

// ********************* Types *******************************

class SgArrayType: public SgType{
  // A new array type is generated for each array.
  // variant == T_ARRAY
public:
   SgArrayType(SgType &base_type);
   int dimension();
   SgExpression *sizeInDim(int i);
   void addDimension(SgExpression *e);
   SgExpression * getDimList();
   SgType * baseType();
   void setBaseType(SgType &bt);
   void addRange(SgExpression &e);
   ~SgArrayType();
};


#if 0
class SgClassType: public SgType{
  // a C struct or Fortran Record, a C++ class, a C Union and a C Enum
  // and a pC++ collection.  note: derived classes are another type.
  // this type is very simple.  it only contains the standard type
  // info from SgType and a pointer to the class declaration stmt
  // and a pointer to the symbol that is the first field in the struct.
  // variant == T_STRUCT, T_ENUM, T_CLASS, T_ENUM, T_COLLECTION
public:
  // why is struct_decl needed. No appropriate field found.
  // assumes that first_field has been declared as
  // FIELD_NAME and the remaining fields have been stringed to it.
  SgClassType(int variant, char *name, SgStatement &struct_decl, int num_fields,
              SgSymbol &first_field);
  SgStatement &structureDecl();
  SgSymbol *firstFieldSymb();
  SgSymbol *fieldSymb(int i);
  ~SgClassType();
};

#endif


class SgPointerType: public SgType{
  // A pointer type contains only one hany bit of information:
  // the base type.
  // can also have a modifier like BIT_CONST BIT_GLOBAL. see SgDescriptType.
  // variant == T_POINTER
public:
  SgPointerType(SgType &base_type);
  SgPointerType(SgType *base_type);
   SgType *baseType();
   int indirection();
   void setIndirection(int);
   int modifierFlag();
   void setModifierFlag(int flag); 
   void setBaseType(SgType &baseType);
   ~SgPointerType();
};


class SgFunctionType: public SgType{
  // Function Types have a returned value type
  // variant == T_FUNCTION
public:
  SgFunctionType(SgType &return_val_type);
  SgType *returnedValue();
  void changeReturnedValue(SgType &rv);
  ~SgFunctionType();
};


class SgReferenceType: public SgType{
  // A reference (&xx in c+=) type contains only one hany bit of information:
  // the base type.
  // variant == T_REFERENCE
public:
   SgReferenceType(SgType &base_type);
   SgType *baseType();
   void setBaseType(SgType &baseType);
   ~SgReferenceType();
   int modifierFlag();
   void setModifierFlag(int flag); 
};

class SgDerivedType: public SgType{
  // for example:  typedef int integer;  go to the symbol table
  // for the base type and Id.
  // variant == T_DERIVED_TYPE
public:
   SgDerivedType(SgSymbol &type_name);
   SgSymbol * typeName();
   ~SgDerivedType();
};

class SgDerivedClassType: public SgType{
  // for example:  typedef int integer;  go to the symbol table
  // for the base type and Id.
  // variant == T_DERIVED_CLASS
public:
   SgDerivedClassType(SgSymbol &type_name);
   SgSymbol *typeName();
   ~SgDerivedClassType();
};

class SgDerivedTemplateType: public SgType{
   // this is the type for a template object: T_DERIVED_TEMPLATE
public:
   SgDerivedTemplateType(SgExpression *arg_vals, SgSymbol *classname);
   SgExpression *argList();
   void addArg(SgExpression *arg);
   int numberOfArgs();
   SgExpression  *arg(int i);
   void setName(SgSymbol &s);
   SgSymbol *typeName(); // the name of the template class.
};

class SgDescriptType: public SgType{
  // for example in C: long volatile int x; 
  // long and volatile are modifiers and there is a descriptor
  // type whose base type is the real type of x.
  // the modifier is an integer with bits set if the modifier
  // holds.
  // the bits are:
  // BIT_SYN, BIT_SHARED, BIT_PRIVATE, BIT_FUTURE, BIT_VIRTUAL, 
  // BIT_, BIT_UNSIGNED, BIT_SIGNED, BIT_LONG, BIT_SHORT,
  // BIT_VOLATILE, BIT_CONST, BIT_TYPEDEF, BIT_EXTERN, BIT_AUTO,
  // BIT_STATIC, BIT_REGISTER, BIT_FRIEND, BIT_GLOBAL, and more.
  //
  // variant = T_DESCRIPT
public:
   SgDescriptType(SgType &base_type, int bit_flag);
   int modifierFlag();
   void setModifierFlag(int flag);
   ~SgDescriptType();
};

class SgDerivedCollectionType: public SgType{
  // for example:
  // Collection DistributedArray {body1} ;
  // class object {body2} ;
  // DistributedArray<object>  X;
  // X is of type with variant = T_DERIVED_COLLECTION
public:
   SgDerivedCollectionType(SgSymbol &s, SgType &t);
   SgType *elementClass();
   void setElementClass(SgType &ty);
   SgSymbol *collectionName();
   SgStatement *createCollectionWithElemType();
   ~SgDerivedCollectionType();
};

// following are added by liyu
class SgDistributeAttrExp : public SgExpression
{
        // for variant=DISTRIBUTE_OP
  inline SgDistributeAttrExp(PTR_LLND ll);
};

class SgAlignAttrExp : public SgExpression
{
        // for variant=ALIGN_ATT
  inline SgAlignAttrExp(PTR_LLND ll);
};

class SgDataDistExp : public SgExpression
{
        // for variant=BLOCK_OP,CYCLIC_OP,COLON_OP,STAR_OP
  inline SgDataDistExp(PTR_LLND ll);
};

// end of liyu addition

// Class definition ends; inline definitions begin

#ifdef USER

SgType *SgTypeInt();
SgType *SgTypeChar();
SgType *SgTypeFloat();
SgType *SgTypeDouble();
SgType *SgTypeVoid();
SgType *SgTypeBool();
SgType *SgTypeDefault();

SgUnaryExp & SgDerefOp(SgExpression &e);
SgUnaryExp & SgAddrOp(SgExpression &e);
SgUnaryExp & SgUMinusOp(SgExpression &e);
SgUnaryExp & SgUPlusOp(SgExpression &e);
SgUnaryExp & SgPrePlusPlusOp(SgExpression &e);
SgUnaryExp & SgPreMinusMinusOp(SgExpression &e);
SgUnaryExp & SgPostPlusPlusOp(SgExpression &e);
SgUnaryExp & SgPostMinusMinusOp(SgExpression &e);
SgUnaryExp & SgBitCompfOp(SgExpression &e);
SgUnaryExp & SgNotOp(SgExpression &e);
SgUnaryExp & SgSizeOfOp(SgExpression &e);
SgUnaryExp & makeAnUnaryExpression(int code,PTR_LLND ll1);


SgValueExp *             isSgValueExp(SgExpression *pt);
SgKeywordValExp *        isSgKeywordValExp(SgExpression *pt);
SgUnaryExp *             isSgUnaryExp(SgExpression *pt);
SgCastExp *              isSgCastExp(SgExpression *pt);
SgDeleteExp *            isSgDeleteExp(SgExpression *pt);
SgNewExp *               isSgNewExp(SgExpression *pt);
SgExprIfExp *            isSgExprIfExp(SgExpression *pt);
SgFunctionCallExp *      isSgFunctionCallExp(SgExpression *pt);
SgFuncPntrExp *          isSgFuncPntrExp(SgExpression *pt);
SgExprListExp *          isSgExprListExp(SgExpression *pt);
SgRefExp *               isSgRefExp (SgExpression *pt);
SgVarRefExp *            isSgVarRefExp (SgExpression *pt);
SgThisExp *              isSgThisExp (SgExpression *pt);
SgArrayRefExp *          isSgArrayRefExp (SgExpression *pt);
SgPntrArrRefExp *        isSgPntrArrRefExp(SgExpression *pt);
SgPointerDerefExp *      isSgPointerDerefExp (SgExpression *pt);
SgRecordRefExp *         isSgRecordRefExp (SgExpression *pt);
SgStructConstExp*        isSgStructConstExp (SgExpression *pt);
SgConstExp*              isSgConstExp (SgExpression *pt);
SgVecConstExp *          isSgVecConstExp (SgExpression *pt);
SgInitListExp *          isSgInitListExp (SgExpression *pt);
SgObjectListExp *        isSgObjectListExp (SgExpression *pt);
SgAttributeExp *         isSgAttributeExp (SgExpression *pt);
SgKeywordArgExp *        isSgKeywordArgExp (SgExpression *pt);
SgSubscriptExp*          isSgSubscriptExp (SgExpression *pt);
SgUseOnlyExp *           isSgUseOnlyExp (SgExpression *pt);
SgUseRenameExp *         isSgUseRenameExp (SgExpression *pt);
SgSpecPairExp *          isSgSpecPairExp (SgExpression *pt);
SgIOAccessExp *          isSgIOAccessExp (SgExpression *pt);
SgImplicitTypeExp *      isSgImplicitTypeExp (SgExpression *pt);
SgTypeExp *              isSgTypeExp (SgExpression *pt);
SgSeqExp *               isSgSeqExp (SgExpression *pt);
SgStringLengthExp *      isSgStringLengthExp (SgExpression *pt);
SgDefaultExp *           isSgDefaultExp (SgExpression *pt);
SgLabelRefExp *          isSgLabelRefExp (SgExpression *pt);
SgProgHedrStmt *         isSgProgHedrStmt (SgStatement *pt);
SgProcHedrStmt *         isSgProcHedrStmt (SgStatement *pt);
SgFuncHedrStmt *         isSgFuncHedrStmt (SgStatement *pt);
SgClassStmt *            isSgClassStmt (SgStatement *pt);
SgStructStmt *           isSgStructStmt (SgStatement *pt);
SgUnionStmt *            isSgUnionStmt (SgStatement *pt);
SgEnumStmt *             isSgEnumStmt (SgStatement *pt);
SgCollectionStmt *       isSgCollectionStmt (SgStatement *pt);
SgBasicBlockStmt *       isSgBasicBlockStmt (SgStatement *pt);
SgForStmt *              isSgForStmt (SgStatement *pt);
SgWhileStmt *            isSgWhileStmt (SgStatement *pt);
SgDoWhileStmt *          isSgDoWhileStmt (SgStatement *pt);
SgLogIfStmt *            isSgLogIfStmt (SgStatement *pt);
SgIfStmt *               isSgIfStmt (SgStatement *pt);
SgArithIfStmt *          isSgArithIfStmt (SgStatement *pt);
SgWhereStmt *            isSgWhereStmt (SgStatement *pt);
SgWhereBlockStmt *       isSgWhereBlockStmt (SgStatement *pt);
SgSwitchStmt *           isSgSwitchStmt (SgStatement *pt);
SgCaseOptionStmt *       isSgCaseOptionStmt (SgStatement *pt);
SgExecutableStatement *  isSgExecutableStatement (SgStatement *pt);
SgAssignStmt *           isSgAssignStmt (SgStatement *pt);
SgCExpStmt *             isSgCExpStmt (SgStatement *pt);
SgPointerAssignStmt *    isSgPointerAssignStmt (SgStatement *pt);
SgHeapStmt *             isSgHeapStmt (SgStatement *pt);
SgNullifyStmt *          isSgNullifyStmt (SgStatement *pt);
SgContinueStmt *         isSgContinueStmt (SgStatement *pt);
SgControlEndStmt *       isSgControlEndStmt (SgStatement *pt);
SgBreakStmt *            isSgBreakStmt (SgStatement *pt);
SgCycleStmt *            isSgCycleStmt (SgStatement *pt);
SgReturnStmt *           isSgReturnStmt (SgStatement *pt);
SgExitStmt *             isSgExitStmt (SgStatement *pt);
SgGotoStmt *             isSgGotoStmt (SgStatement *pt);
SgLabelListStmt *        isSgLabelListStmt (SgStatement *pt);
SgAssignedGotoStmt *     isSgAssignedGotoStmt (SgStatement *pt);
SgComputedGotoStmt *     isSgComputedGotoStmt (SgStatement *pt);
SgStopOrPauseStmt *      isSgStopOrPauseStmt (SgStatement *pt);
SgCallStmt*              isSgCallStmt (SgStatement *pt);
SgProsHedrStmt *	 isSgProsHedrStmt (SgStatement *pt); /* Fortran M */
SgProcessDoStmt *        isSgProcessDoStmt (SgStatement *pt);  /* Fortran M */
SgProsCallStmt*          isSgProsCallStmt (SgStatement *pt); /* Fortran M */
SgProsCallLctn*          isSgProsCallLctn (SgStatement *pt); /* Fortran M */
SgProsCallSubm*          isSgProsCallSubm (SgStatement *pt); /* Fortran M */
SgInportStmt *           isSgInportStmt (SgStatement *pt); /* Fortran M */
SgOutportStmt *          isSgOutportStmt (SgStatement *pt); /* Fortran M */
SgIntentStmt *		 isSgIntentStmt (SgStatement *pt); /* Fortran M */
SgChannelStmt *          isSgChannelStmt (SgStatement *pt); /* Fortran M */
SgMergerStmt *           isSgMergerStmt (SgStatement *pt); /* Fortran M */
SgMoveportStmt *         isSgMoveportStmt (SgStatement *pt); /* Fortran M */
SgSendStmt *             isSgSendStmt (SgStatement *pt); /* Fortran M */
SgReceiveStmt *          isSgReceiveStmt (SgStatement *pt); /* Fortran M */
SgEndchannelStmt *       isSgEndchannelStmt (SgStatement *pt); /* Fortran M */
SgProbeStmt *            isSgProbeStmt (SgStatement *pt); /* Fortran M */
SgProcessorsRefExp *	 isSgProcessorsRefExp(SgExpression *pt); /* Fortran M */
SgPortTypeExp *          isSgPortTypeExp (SgExpression *pt); /* Fortran M */ 
SgInportExp *            isSgInportExp (SgExpression *pt); /* Fortran M */
SgOutportExp *           isSgOutportExp (SgExpression *pt); /* Fortran M */
SgFromportExp *          isSgFromportExp (SgExpression *pt); /* Fortran M */
SgToportExp *            isSgToportExp (SgExpression *pt); /* Fortran M */
SgIO_statStoreExp *      isSgIO_statStoreExp (SgExpression *pt); /* Fortran M */
SgEmptyStoreExp *        isSgEmptyStoreExp (SgExpression *pt); /* Fortran M */
SgErrLabelExp *          isSgErrLabelExp (SgExpression *pt); /* Fortran M */
SgEndLabelExp *          isSgEndLabelExp (SgExpression *pt); /* Fortran M */
SgDataImpliedDoExp *     isSgDataImpliedDoExp (SgExpression *pt);/* Fortran M */
SgDataEltExp *           isSgDataEltExp (SgExpression *pt); /* Fortran M */
SgDataSubsExp *          isSgDataSubsExp (SgExpression *pt); /* Fortran M */
SgDataRangeExp *         isSgDataRangeExp (SgExpression *pt); /* Fortran M */
SgIconExprExp *          isSgIconExprExp (SgExpression *pt); /* Fortran M */
SgIOStmt *               isSgIOStmt (SgStatement *pt);
SgInputOutputStmt *      isSgInputOutputStmt (SgStatement *pt);
SgIOControlStmt *        isSgIOControlStmt (SgStatement *pt);
SgDeclarationStatement * isSgDeclarationStatement (SgStatement *pt);
SgVarDeclStmt *          isSgVarDeclStmt (SgStatement *pt);
SgVarListDeclStmt *      isSgVarListDeclStmt (SgStatement *pt);
SgStructureDeclStmt *    isSgStructureDeclStmt (SgStatement *pt);
SgNestedVarListDeclStmt* isSgNestedVarListDeclStmt (SgStatement *pt);
SgParameterStmt *        isSgParameterStmt (SgStatement *pt);
SgImplicitStmt *         isSgImplicitStmt (SgStatement *pt);
SgVariableSymb *         isSgVariableSymb (SgSymbol *pt);
SgConstantSymb *         isSgConstantSymb (SgSymbol *pt);
SgFunctionSymb *         isSgFunctionSymb (SgSymbol *pt);
SgMemberFuncSymb *       isSgMemberFuncSymb (SgSymbol *pt);
SgFieldSymb *            isSgFieldSymb (SgSymbol *pt);
SgClassSymb *            isSgClassSymb (SgSymbol *pt);
SgLabelSymb *            isSgLabelSymb (SgSymbol *pt);
SgLabelVarSymb *         isSgLabelVarSymb (SgSymbol *pt);
SgExternalSymb *         isSgExternalSymb (SgSymbol *pt);
SgConstructSymb *        isSgConstructSymb (SgSymbol *pt);
SgInterfaceSymb *        isSgInterfaceSymb (SgSymbol *pt);
SgModuleSymb *           isSgModuleSymb (SgSymbol *pt);
SgArrayType *            isSgArrayType (SgType *pt);
SgPointerType *          isSgPointerType (SgType *pt);
SgFunctionType *         isSgFunctionType (SgType *pt);
SgReferenceType *        isSgReferenceType (SgType *pt);
SgDerivedType *          isSgDerivedType (SgType *pt);
SgDerivedClassType *     isSgDerivedClassType (SgType *pt);
SgDescriptType *         isSgDescriptType (SgType *pt);
SgDerivedCollectionType* isSgDerivedCollectionType (SgType *pt);
#endif

#endif /* ndef LIBSAGEXX_H */
